AppVer = '0.5.1'

try:
    # 防止IDE自动删除 或者整理 import内容
    import io

    from dataclasses import dataclass
    import random
    import queue
    import itertools as ite

    import yaml
    from sortedcontainers import SortedDict  # type:ignore
    from dataclasses import field

    from urllib.parse import parse_qsl

    from dataclasses_json import dataclass_json

    from uuid import uuid4

    from pydantic import BaseModel

    from collections import deque
    import json
    from queue import Queue
    from typing import Deque, cast, Protocol, Sequence

    from urllib.parse import urlparse
    from urllib.parse import parse_qs, urlencode

    import html

    import functools
    # import schedule

    import numpy.typing as npt

    import numpy as np
    from typing import TypeAlias

    MyNDArrayInt64: TypeAlias = npt.NDArray[np.int64]
    MyNDArrayInt16: TypeAlias = npt.NDArray[np.int16]
    MyNDArrayInt8: TypeAlias = npt.NDArray[np.int8]
    MyNDArrayFloat64: TypeAlias = npt.NDArray[np.float64]
    MyNDArrayFloat32: TypeAlias = npt.NDArray[np.float32]

    from attr import define, field as afield, asdict

    from dknovautils import *
    from dknovautils import (
        AT,
        iprint_debug,
        iprint_info,
        iprint_warn,
        iprint,
        DkFile,
        join,
        List,
        Dict,
        Tuple,
        Set,
        Iterator,
        time,
        threading,
        Path,
        np,
        os,
        shutil,
        Any,
    )

    from dknovautils.dkprocess import DKProcessUtil

    run_simple_a = DKProcessUtil.run_simple_a
    run_simple_b = DKProcessUtil.run_simple_b

    from .common_ctts_ytb import *

finally:
    pass



from __future__ import annotations

from takahom.common.down_request import DownRequest
from takahom.common.common_etc import limit_memory, AT2
from takahom.server.server_base_c_download import IServerBaseC
from takahom.server.server_config import ServerConfig, build_config_from_args
from takahom.server.server_index import DLServerIndex

from takahom.server.server_ytb_util import *

from takahom.server.server_base_a import IServerBaseA, dec_my_log_error_builder, Source_Demo, build_system_info
from attr import define, field as afield, asdict


def check_ydltest_url(server: IServerBaseA) -> None:
    """
    重要目录 sinput目录等 完成扫描 和检查
    检查scan目录都是存在的
    测试ytb url可以正常下载 ,info log begin end error

    """
    self = server

    if True:
        url = ytb_url_demo_ydltest
        res = 720
        req = build_demo_request(server, url, res)
        iprint_info(f'try to test and download {url} begin')
        res = server.handle_request(req)
        if not res.ok:
            AT.fail(f'err7611 test and download {url} failure')
        iprint_info(f'test and download {url} ok')
        # clear dir
        dir_scache = self.subdir('scache')
        CommonUtil.clear_dir(dir_scache)

    pass


def build_demo_request(appctx: IServerBaseA, url: str, res: int) -> DownRequest:
    ytbitem = YtbUtil.parse_ytbitem_from_url(url)
    assert ytbitem

    req = DownRequest(url=ytbitem.get_url(), ytb_resolution=res)
    req.init()
    req.ytb_item = cast(YtbItem, ytbitem)
    req.ytb_ytbid = cast(YtbId, YtbUtil.parse_ytbid_from_url(url))
    req.domain = ytb_domain_name
    req.resource_key = req.ytb_ytbid.to_str_with_prefix()
    req.proxy = appctx.socks_proxy
    req.cmd = appctx.ytb_cmd_path
    req.source = Source_Demo
    return req


@dec_my_log_error_builder(message="Error on tgt_main: {e!r}")  # type:ignore
def tgt_main_download_queue(self: IServerBaseC) -> None:
    for cnt in ite.count(0):

        while CommonUtil.dir_has_filename(self.input_dir, "exit.cmd"):
            iprint_info("step55341 file exit.cmd found. sleep.")
            time.sleep(60)

        # 可以用文件让循环主动break 不能exit只好改成continue 这样可以循环打印日志
        if False and CommonUtil.dir_has_filename(self.input_dir, "exit.cmd"):
            # do nothing
            iprint_warn("step55341 file exit.cmd found. exit application.")
            # 为什么 sys.exit 不能生效?
            # sys.exit()
            pass
        else:
            iprint_debug(f"main read queue begin")
            self.thread_main_download(cnt=cnt)
            iprint_debug(f"main read queue end")

        sp = 30 if AT2.is_profile_prod() else 3
        time.sleep(sp)


@dec_my_log_error_builder(message="Error on tgt_remove_toomuch_data: {e!r}")  # type:ignore
def tgt_remove_toomuch_data(self: IServerBaseC) -> None:
    SECS = self.s7z_keep_secs
    if not SECS:
        return
    AT.assert_(SECS > 60 * 10, 'err19969')

    for _ in range(1):
        RX = 0
        for dkf in DkFile2.oswalk_simple_a(self.subdir('s7z'), only_file=True):
            rv = int(CommonUtil.parse_filename_dict(dkf).get('r', '0'))
            RX = max(RX, rv)

        if RX:
            from os.path import getmtime
            cnt = 0
            for sdir in ['s7z', 'sresp']:
                for dkf in DkFile2.oswalk_simple_a(self.subdir(sdir), only_file=True):
                    rv = int(CommonUtil.parse_filename_dict(dkf).get('r', '0'))
                    now = AT.fepochSecs()
                    RN = 3
                    if rv and rv + RN < RX and dkf.is_file() \
                            and now - getmtime(dkf.path) > SECS:
                        os.remove(dkf.path)
                        iprint_debug(f'step28568 删除过多数据 {dkf}')
                        cnt += 1
            if cnt:
                iprint_info(f'step45645 remove useless data {cnt}')


@dec_my_log_error_builder(message="Error on tgt_email_log: {e!r}")  # type:ignore
def tgt_email_log(self: IServerBaseA, write_email: bool = True) -> str:
    infodc = tgt_job_log_5mins(self, _print=False)
    infodc['sysinfo'] = asdict(build_system_info(self.work_dir))
    infodc['conf'] = build_config_from_args(self.args).copy_without_secret()

    json_string = json.dumps(infodc, ensure_ascii=False, indent=4)

    if write_email:
        setc_dir = self.subdir("setc")
        fname = f"email_{AT.sdf_logger_format_datetime(noColon=True)}.json"
        Path(join(setc_dir.pathstr, fname)).write_text(json_string)
        iprint_info(f'write email log file {fname}')

    return json_string


_demo_v_state: List[str] = ['ydltest', 'list', 'short']


def f_try_to_add_demo(self: IServerBaseC) -> None:
    v_state = _demo_v_state
    iprint_debug(f"demo req {v_state}")
    if not v_state:
        return

    if not self.req_queue.empty():
        iprint_debug("queue is not empty. wait awhile.")
        return

    assert self.ytb_cmd_path

    id = v_state[-1]
    if id == 'ydltest':
        url = ytb_url_demo_ydltest
        res = 720
        req = build_demo_request(self, url, res)

        reqs = [req]

    elif id == 'list':
        url = "https://www.youtube.com/watch?v=csebxC-oKn4&list=PLtpBIq5yTxAUrB1-jkE9uNQludEzbhLmL"
        url = "https://www.youtube.com/watch?list=PLtpBIq5yTxAUrB1-jkE9uNQludEzbhLmL"
        # 没有v值的url 是不能下载list的
        url_歌曲30首 = "https://www.youtube.com/watch?list=RDf-Neutoz40g"
        url_songs = "https://www.youtube.com/watch?v=f-Neutoz40g&list=RDf-Neutoz40g&_maxitems_=3"

        url = url_songs
        res = 480
        req = build_demo_request(self, url, res)
        reqs = [req]

    elif id == 'short':
        url_shorta = "https://www.youtube.com/shorts/6fC95hU3puU"

        url = url_shorta
        res = 480
        req = build_demo_request(self, url, res)
        reqs = [req]

    else:
        reqs = []

    if not reqs:
        iprint_debug(f"no new url")
    else:
        iprint_debug(f"find new urls total:{len(reqs)}")
        req = reqs[0]
        self.put_req(req)
        v_state.pop(-1)
        time.sleep(3.0)


def f_try_to_add_ytb_scan_dir(self: IServerBaseC) -> None:
    app_ctx = self
    assert self.req_size_max >= 1, 'err27326'
    nps = self.req_size_max - len(self.req_queue_list)
    if nps <= 0:
        iprint_debug("queue is not empty. wait awhile.")
        return

    # 排除 excludecodes， output_dir中的重复，选择一个url

    exitids: Set[YtbId] = set(ytb_scan_ytbids_excludes_file(app_ctx.input_dir))
    exitids.update(ytb_scan_output_ytbids_scan_dirs(
        scan_dirs=app_ctx.scan_dirs,
        output_dirs=[app_ctx.output_dir]))

    if req_run := self.req_running:
        exitids.add(req_run.ytb_ytbid)

    exitids.update(item.para_value_to_ytbid() for item in self.req_queue_list)

    assert app_ctx.ytb_cmd_path

    reqs = [
        req
        for req in ytb_parse_input_dir_reqs(
            input_dir=app_ctx.input_dir,
            socks_proxy=app_ctx.socks_proxy,
            cmd_path=app_ctx.ytb_cmd_path,
        )
        if req.ytb_ytbid not in exitids
    ]

    self.last_new_urls = len(reqs)
    if not reqs:
        iprint_debug(f"no new url. exitids:{len(exitids)}")
    else:
        iprint_debug(f"find new urls total:{len(reqs)} exitids:{len(exitids)}")
        for i in range(nps):
            req: DownRequest = reqs[i]
            self.put_req(req)

    time.sleep(3.0)


def thread_auto_req_job(self: IServerBaseC, scnt: int) -> None:
    """
    周期调用delay是60秒 scnt大概60秒增加一次

    """

    dir_scache = self.subdir("scache")

    def f_add_test() -> None:
        r = random.randint(1, 10 ** 5)
        req = DownRequest(
            url=f"http://test/abc?r={r}",
            data_dir=dir_scache.pathstr,
            uuid=AT2.gen_uuid_str(),
            start_epoch=AT.fepochSecs(),
        )
        req.init()

        self.req_queue.put(req)

    def f_add_local(n: int) -> None:
        DownRequest(url="http://test")

        pass

    def f_add_curl() -> None:
        r = random.randint(1, 10 ** 5)
        req = DownRequest(
            url=CommonUtil.test_random_github_url(),
            data_dir=dir_scache.pathstr,
            uuid=AT2.gen_uuid_str(),
            start_epoch=AT.fepochSecs(),
        )
        req.init()

        self.req_queue.put(req)

    def f_add_ytb() -> None:
        r = random.randint(1, 10 ** 5)
        req = DownRequest(
            url=CommonUtil.special_check_urls_ytb[0] + f"&r={r}",
            data_dir=dir_scache.pathstr,
        )
        req.init()

        self.req_queue.put(req)

    base = 60 if AT2.is_profile_prod() else 3
    if scnt % base == 0:
        f_add_test()

    base = 60 if AT2.is_profile_prod() else 3
    if scnt % base == 0:
        f_add_curl()

    base = 60 if AT2.is_profile_prod() else 3
    if scnt % base == 0:
        f_add_ytb()


@dec_my_log_error_builder(message="Error on tgt_job_log_5mins: {e!r}")  # type:ignore
def tgt_job_log_5mins(self: IServerBaseC, _print: bool = True) -> Dict[str, Any]:
    """
    每5分钟调用一次

    info级别 打印一次运行状态
        统计主要路径的运行次数
        异常次数
    报告sdownload目录大小
    """

    rtss = [f'{ts.rate // 1000}' for ts in self.rate_timespans[-13:]]

    source = self.req_running.source if self.req_running else ''
    infodc = {
        'events': SortedDict(self.stt_cnts),
        'downloading_url': self.downloading_url,
        'downloading_begin': self.downloading_begin,
        'downloading_size': f'{self.scache_size:,} bytes',
        'downloading_file': source,
        'appending_urls': ', '.join(x.url for x in self.req_queue_list),
        'downloaded_total_bytes': f'{self.downloaded_bytes:,} bytes',
        'more_urls': self.last_new_urls,
        'log_errs': AT._AstErrorCnt_,
        'log_warns': AT._AstWarnCnt_,
        'network_rates kb/s': ', '.join(rtss),
        'app_start': self.start_app_epoch_str,
        'now': AT.sdf_logger_format_datetime(),

    }

    if _print:
        jstr = json.dumps(infodc, ensure_ascii=False, indent=4)
        iprint_info(jstr)
    return infodc



from __future__ import annotations

from takahom.common.common_etc import limit_memory, AT2
from takahom.server.server_base_a import IServerBaseA, copy_ytbtxts_files_to_dest_input_dir
from takahom.server.server_base_web import IServerBaseWeb
from takahom.server.server_base_c_download import IServerBaseC
from takahom.server.server_base_d import build_system_info, check_ydltest_url, build_demo_request, \
    tgt_main_download_queue, \
    tgt_remove_toomuch_data, tgt_email_log, f_try_to_add_demo, \
    f_try_to_add_ytb_scan_dir, \
    tgt_job_log_5mins
from takahom.server.server_base_e import INetCalulator, DirSizeInfo, TimeSpan
from takahom.server.server_config import ServerConfig, build_config_from_args, test001
from takahom.server.server_index import DLServerIndex

from takahom.server.server_base_a import dec_my_log_error_builder

from takahom.server.server_ytb_util import *


class DownloadServer(IServerBaseC, INetCalulator):
    subdirs = ["scache", "sinput", "sdownload", "s7z", "sresp", "stmp", "setc", "slog"]

    def __init__(
            self,
            args: Any,
            *,
            sconf: ServerConfig,
    ) -> None:
        self.start_app_epoch_str = AT.sdf_logger_format_datetime()
        self.is_mode_demo: bool = args.mode == "demo"
        self.is_mode_run: bool = args.mode == "run"

        self.sconf = sconf
        self.args: Any = args

        self.fkey = sconf.file_key
        self.secret = sconf.secret
        self.port = sconf.port

        self.work_dir = DkFile(sconf.work_dir)
        self.p7z_volume = sconf.p7z_volume
        self.p7z_level = sconf.p7z_level

        self.cfg_fid = sconf.cfg_fid
        self.cfg_rid = sconf.cfg_rid

        # 30天之内 同样的url不能重复下载
        self.url_duplicate_check_days = sconf.url_duplicate_check_days

        # 网速检测 低于该网速则主动停止传输
        self.v130_limit_break = sconf.v130_limit_break

        self.rate_limits: List[int] = sconf.cal_rate_limits(sconf.rate_limits)
        CommonUtil.rate_limit(self.rate_limits)
        iprint_info(f'rate limites: {self.rate_limits}')

        self.socks_proxy: str = sconf.network_proxy
        self.ytb_cmd_path: str = sconf.ytb_cmd_path

        self.s7z_keep_secs: int = sconf.s7z_keep_secs

        self.scan_dirs: List[DkFile] = sconf.cal_scan_dirs()
        iprint_info(f'scan_dirs: {[dkf.pathstr for dkf in self.scan_dirs]}')

        if vStep1OK := True:
            AT.assert_(
                self.work_dir.exists() and self.work_dir.is_dir(),
                f"err51563 dir not exists {self.work_dir.path}",
            )

            for d in self.subdirs:
                os.makedirs(join(self.work_dir.path, d), exist_ok=True)

            iprint_info('start app step1 ok')

        self.stt_cnts: Dict[str, int] = dict()
        self.dir_size_infos: List[DirSizeInfo] = []
        self.rate_timespans: List[TimeSpan] = []
        self.downloading_url: str | None = None
        self.downloading_begin: str | None = None
        self.scache_size: int = 0
        self.downloaded_bytes: int = 0

        self._empty_postfix = ".empty"
        self.fid = 10 ** 7
        self.state_index: DLServerIndex
        dir_sinput = self.subdir("sinput")
        dir_sdownload = self.subdir("sdownload")

        self.input_dir = dir_sinput
        self.output_dir = dir_sdownload
        self.sresp_dir = self.subdir("sresp")

        # 没用到 可以删除
        self.allow_duplicate_url = False
        if AT2.is_profile_dev():
            self.allow_duplicate_url = False

        # 保证同时添加和移除
        self.req_queue: Queue[DownRequest] = Queue()
        self.req_queue_list: List[YtbItem] = []
        self.req_size_max: int = 3
        self.req_running: DownRequest | None = None
        self.sresp_fids: MyNDArrayInt64 | None = None
        self.sresp_fnames: Dict[int, str] = {}
        self.last_new_urls: int = 0

        if vStep2OK := True:
            sinfo = build_system_info(self.work_dir)
            iprint_info('start app step2 ok')
            AT.assert_(self.input_dir.exists(), f"dir doesn't exist. {self.input_dir.pathstr}")

            def dirok(d: DkFile) -> bool:
                return d.exists() and d.is_dir()

            dsok = all(dirok(d) for d in self.scan_dirs)
            AT.assert_(dsok, f'err98240 some directory is bad. {self.scan_dirs}')

    @dec_my_log_error_builder(message="Error on thread_scan_input_req_scan_dir: {e!r}")  # type:ignore
    def thread_scan_input_req_scan_dir(self) -> None:
        """
        """
        if self.is_mode_run:
            f_try_to_add_ytb_scan_dir(self)
        elif self.is_mode_demo:
            f_try_to_add_demo(self)
        else:
            AT.never()

    def server_init_stateindex(self) -> None:
        """
        """

        _index_file_data = self.read_index_file()

        if _index_file_data is None:
            ss = DLServerIndex()
            ss.results = []
            _index_file_data = ss
        self.state_index = _index_file_data

        while len(self.state_index.results) > self.sconf.state_res_max_size:
            self.state_index.results.pop(0)

        def f_maxfidfile_sresp(sresp_dir: DkFile) -> DkFile | None:
            fid = 0
            file = None
            for d in os.scandir(sresp_dir.pathstr):
                dkf = DkFile(join(sresp_dir.pathstr, d.name))
                dc = CommonUtil.parse_filename_dict(dkf)
                f = int(dc["fid"])
                assert f > 0
                if fid == 0 or f > fid:
                    fid = f
                    file = dkf

            return file

        _maxfidfile_sresp = f_maxfidfile_sresp(self.sresp_dir)
        _maxfidfile_sresp_dc = (
            None
            if not _maxfidfile_sresp
            else CommonUtil.parse_filename_dict(_maxfidfile_sresp)
        )

        if True:
            AT.assert_(self.cfg_rid >= 10 ** 5)
            if _maxfidfile_sresp:
                assert _maxfidfile_sresp_dc
                file_maxid = int(_maxfidfile_sresp_dc["r"])
            else:
                file_maxid = 0

            if _index_file_data:
                indexfile_id = _index_file_data.rid
            else:
                indexfile_id = 0

            self.state_index.rid = max(self.cfg_fid, file_maxid + 1, indexfile_id)

    def start(self) -> None:

        self.server_init_stateindex()

        copy_ytbtxts_files_to_dest_input_dir(src=DkFile('./docs'),
                                             dest=self.subdir('sinput'))

        if vStep3OK := True:
            # 扫描 sdownload scan_dirs 并完成中文简繁体转换
            for dkf in [self.output_dir]:
                CommonUtil.chinese_convert_directory(dkf, cvt=self.chinese_cvt_enum)

            for dkf in self.scan_dirs:
                CommonUtil.chinese_convert_directory(dkf, cvt=self.chinese_cvt_enum)

            check_ydltest_url(self)
            iprint_info('start app step3 ok')

        f_try_to_add_ytb_scan_dir(self)

        thread = threading.Thread(target=tgt_main_download_queue,
                                  args=(self,),
                                  daemon=False  # main thread may end at once
                                  )
        thread.start()

        schedule.every(5).minutes.do(lambda: self.thread_scan_input_req_scan_dir())

        mins = 60 if AT2.is_profile_prod() else 3
        schedule.every(mins).minutes.do(lambda: tgt_remove_toomuch_data(self))

        mins = 60 if AT2.is_profile_prod() else 3
        schedule.every(mins).minutes.do(lambda: tgt_email_log(self))

        mins = 5 if AT2.is_profile_prod() else 1
        schedule.every(mins).minutes.do(lambda: tgt_job_log_5mins(self))

        dir_scache = self.subdir('scache')
        schedule.every(60).seconds.do(lambda: self.tgt_update_sdownload_size(dkf=dir_scache))

        def th_run_schedule() -> None:
            while True:
                schedule.run_pending()
                time.sleep(1)

        thread = threading.Thread(target=th_run_schedule, args=(), daemon=True)
        thread.start()

        if vStep4OK := True:
            iprint_info(f'start app step4 ok')

        if True:
            self.start_web_server()


def main(args: Any) -> None:
    AT2._app_profile = args.profile
    sc = build_config_from_args(args)

    AT.assert_(AT.is_linux, 'err15238 only support linux os')

    if not AT2.is_profile_prod():
        test001()

    max_memory = 1024 * 1024 * 100
    limit_memory(max_memory)

    server = DownloadServer(args, sconf=sc)
    AT2.init_logging_v1(DkFile('.'), log_dir=server.subdir('slog'))
    iprint_info(f"server start version:{AppVer}")
    iprint_info(f"config file, {sc.conf_file}: {sc.copy_without_secret()}")
    time.sleep(0.5)

    server.start()


"""



"""



from __future__ import annotations

import hashlib
import itertools as ite
import os
import shutil
from os.path import join
from pathlib import Path
from typing import Iterator, Tuple, Callable, Any, Iterable
from uuid import uuid4

import yaml
from dknovautils import classproperty, AT, DkFile, iprint, iprint_warn, LLevel, run_simple_a
from pyrate_limiter import Limiter, Rate, Duration


def eat_for(ps: Iterable[Any]) -> None:
    for p in ps:
        pass


class AT2:
    _profile_prod = "prod"
    _profile_test = "test"
    _profile_dev = "dev"
    _profiles_list = [_profile_prod, _profile_test, _profile_dev]

    _app_profile: str = ""

    @classproperty
    def app_profile(cls) -> str:
        AT.assert_(cls._app_profile in cls._profiles_list, 'err46501 bad app profile')
        return cls._app_profile

    @classmethod
    def is_profile_prod(cls) -> bool:
        return bool(cls._profile_prod == cls.app_profile)

    @classmethod
    def is_profile_dev(cls) -> bool:
        return bool(cls._profile_dev == cls.app_profile)

    @classmethod
    def is_profile_test(cls) -> bool:
        return bool(cls._profile_test == cls.app_profile)

    def _dummy_drop_fun(s: Any) -> None:
        return None

    @classmethod
    def tprint_fun(cls, verbose: bool = True,
                   use_print: bool = False) -> Callable[[Any], None]:

        if not verbose:
            return cls._dummy_drop_fun
        else:
            # 这两个函数可以抽离成2个静态函数 不需要创建变量
            def fn(msg: str) -> None:
                if not use_print:
                    iprint(msg, level=LLevel.Debug)
                else:
                    print(msg)

            return fn

    @classmethod
    def gen_uuid_str(cls) -> str:
        r = uuid4()
        r = str(r)
        return r

    @classmethod
    def md5_hex(cls, bs: bytes) -> str:
        md5 = hashlib.md5()
        md5.update(bs)
        r = md5.hexdigest().lower()
        AT.assert_(len(r) == 32)
        return r

    @classmethod
    def file_md5_hex(cls, file_path: str | Path) -> str:
        """
        计算文件的md5
        :param file_name:
        :return:

        https://www.cnblogs.com/xiaodekaixin/p/11203857.html

        """
        m = hashlib.md5()  # 创建md5对象
        with open(file_path, "rb") as fobj:
            while True:
                data = fobj.read(4096)
                if not data:
                    break
                m.update(data)  # 更新md5对象

        return m.hexdigest().lower()  # 返回md5对象

    @classmethod
    def init_logging_v1(cls, conf_dir: DkFile, *, log_dir: DkFile | None = None) -> None:
        """

        conf_dir 配置文件目录 必须存在
            logging_$profile.yaml

        log_dir 日志文件目录 会自动创建
            可以用特殊字符来替换配置文件中的目录 将 _LOGDIR_ 替换为 日志目录

    可以调用 init_logging_v1 进行初始化
        将读取yaml配置文件进行初始化
        将设置 _innerLoggerFun_ 用来作为 iprint的执行函数
        """
        AT.assert_(conf_dir.is_dir() and conf_dir.exists(), 'err69053')

        LOGDIR = 'logs/'
        AT.assert_(AT2.app_profile, 'err21738 bad profile')
        conf_file = DkFile(join(conf_dir.pathstr, f'logging_{AT2.app_profile}.yaml'))
        cstr = conf_file.path.read_text()
        iprint(f'logging conf_file: {conf_file} len:{len(cstr)}')
        if log_dir:
            AT.assert_(LOGDIR in cstr, f'err12912 cannot find {LOGDIR} in yaml ')
            cstr = cstr.replace(LOGDIR, log_dir.pathstr + '/')
            os.makedirs(log_dir.path, exist_ok=True)
        conf = yaml.safe_load(cstr)
        import logging.config as log_config
        log_config.dictConfig(conf)
        AT.log_loggerFun(initInnerLoggerFun=True)

    @classmethod
    def ln_filter(cls, lns: Iterable[str], map_opt: Callable[[str], str | None]) -> Iterator[str]:
        # 返回的是 optional 这是大象符号引入之后 更方便的形式
        for ln in lns:
            r = map_opt(ln.strip())
            if r is not None:
                yield r
            else:
                continue

    @classmethod
    def ln_filter_startswith(cls, lns: Iterable[str], stt: Tuple[str, ...]) -> Iterator[str]:
        return cls.ln_filter(lns, map_opt=lambda ln: ln if any(ln.startswith(e) for e in stt) else None)

    @classmethod
    def ln_filter_not_startswith(cls, lns: Iterable[str], stt: Tuple[str, ...]) -> Iterator[str]:
        return cls.ln_filter(lns, map_opt=lambda ln: ln if not any(ln.startswith(e) for e in stt) else None)


class DkFile2:

    @staticmethod
    def clear_dir(tmp_dir: DkFile) -> None:
        # 清空tmp目录
        if AT.is_windows:
            shutil.rmtree(tmp_dir.pathstr)
            os.mkdir(tmp_dir.pathstr)
        else:
            run_simple_a(f'rm -rf "{tmp_dir.pathstr}"/* ')

    @classmethod
    def oswalk_simple_a(cls,
                        topdir: DkFile,
                        topdown: bool = True,
                        only_file: bool = False,
                        ) -> Iterator[DkFile]:
        i = 0
        for _i, (root, dirs, files) in enumerate(os.walk(topdir.path, topdown=topdown)):
            for d in ite.chain(dirs, files):
                # assert isinstance(d, str) and isinstance(root, Path), f'{d} {root}'
                dkf = DkFile(join(root, d))
                if only_file and not dkf.is_file():
                    continue

                yield dkf
                i += 1

    @classmethod
    def join_path(cls, dkf: DkFile, filename: str | Path) -> DkFile:
        return DkFile(join(dkf.path, filename))

    @classmethod
    def basename_without_extension(cls, dkf: DkFile) -> str:
        file_extension = dkf.path.suffix
        if file_extension:
            return dkf.basename[:-len(file_extension)]
        else:
            return dkf.basename


def limit_memory(maxsize: int) -> None:
    # import resource
    # soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    # resource.setrlimit(resource.RLIMIT_AS, (maxsize, hard))
    pass


def f_simple_limiter(vlmt: int) -> Tuple[Limiter, int]:
    """
    v 每秒网速限制 byte 1kb/s = 1000
    因为tksize=v/10 而且一般情况 tksize一般不能太大 比如 不要超过1M
    """
    assert vlmt >= 1000, f"err41849 bad v. vlmt >=1000 "

    def f_n() -> int:
        # 防止 tksize太大 不便于流量控制的效果
        n: float = 10.0
        while True:
            tks = vlmt / n
            if tks < 1000 * 100:
                assert n >= 10
                return int(n)
            else:
                n *= 1.2

    kn = f_n()
    tksize = int(vlmt / kn)
    rate = Rate(kn, Duration.SECOND * 1)
    limiter = Limiter(rate, raise_when_fail=False, max_delay=Duration.SECOND * 10)
    return limiter, tksize



from __future__ import annotations

import attr.exceptions

from takahom.common.common_util import *
from takahom.common.common_etc import DkFile2
from takahom.server.server_base_a import IServerBaseA, dec_my_log_error_builder


@define(slots=True, frozen=True, kw_only=True)
class DirSizeInfo:
    ctime_ms: int  # epoch millis
    size_n: int  # 该时刻的sdownload目录的大小 字节数


TS_LEN: int = 1000 * 60 * 5  # 5mins长度的毫秒值


@define(slots=True, frozen=True, kw_only=True)
class TimeSpan:
    start: int  # ms
    end: int  # exclusive
    rate: int = -1  # Bytes/sec B/s

    def next_span(self, n: int = 1, rate: int = -1) -> TimeSpan:
        r = TimeSpan(start=self.start + n * TS_LEN,
                     end=self.end + n * TS_LEN,
                     rate=rate)
        return r


class INetCalulator(Protocol):
    dir_size_infos: List[DirSizeInfo]
    rate_timespans: List[TimeSpan]
    scache_size: int
    downloaded_bytes: int

    @dec_my_log_error_builder(message="Error on tgt_job_update_sdownload_size: {e!r}")  # type:ignore
    def tgt_update_sdownload_size(self, dkf: DkFile) -> None:
        # 60s调用一次
        cmd = f'du "{dkf.pathstr}" -d0 --block-size=1'
        rc, out, err = run_simple_b(cmd)
        self.scache_size = int(out.strip().split()[0])
        self.f_add_sizeinfo(DirSizeInfo(ctime_ms=AT.fepochMillis(), size_n=self.scache_size))
        self.tgt_job_update_netrates()

    def f_add_sizeinfo(self, dsi: DirSizeInfo) -> None:
        """
        大概是每40s - 60s的时间间隔产生一个数据
        """
        AT.assert_(dsi.ctime_ms > 0 and dsi.size_n >= 0, 'err87643')
        dus = self.dir_size_infos
        if dus:
            AT.assert_(dus[-1].ctime_ms + 10 < dsi.ctime_ms, 'err14647')
            diff = dsi.size_n - dus[-1].size_n
            if diff > 0:
                self.downloaded_bytes += diff

        dus.append(dsi)
        while len(dus) >= 50:
            dus.pop(0)

    def tgt_job_update_netrates(self) -> None:
        """
        本函数大概是60s的周期被调用

        返回按照每N=5分钟汇总计算的平均网速 KBytes/secons KB/s

        数据参数中
            dus的数据也是按照时间的升序排列 大概是每40s - 60s的时间间隔产生一个数据
            因为是同一个thread内部 所以不需要 临界区了

        最多保存100个5分钟的数据 如果某分钟内无数据 则对应的数据为零
        时间最近的数据排在后面 按照时间升序排列

        约束和校验
            dus中时间值是严格升序的 否则异常 间隔>10s size_n>=0 len(dus)<100 不需要检查已经保证
            rts 也是严格升序的 而且是连续的 如果 len(rts)>100 直接pop(0)即可

        算法
            fa： 输入 时刻点集，返回 包括所有时间的连续的区间遍历器（不会超出这个时间的范围，就是说，最早和最晚的区间，一定包含某个时间点）
                从 最早时刻 一直连续到 最晚的时刻 的区间集合
            fb 两个参数，时间点和区间 返回两者之间的关系 -1表示时间点在区间之前，0表示在区间内 1表示在区间后

            fc 两个参数，区间r和时间点p，返回 区间r+1 ... 区间x 的遍历器。 区间x是所有<p的最大的一个区间
                在fa的基础上实现即可
                所以其实是参数的全开区间的范围内的遍历

            fd 两个参数 区间r和时间点集ps 计算区间内的网速数值
                ps是时间升序排列的
                sum(diff for a,b in pairwise(ps)
                    if b in span_r
                    if (diff := b.sizen -a.sizen)>0
                    )



        """

        L = TS_LEN  # 5mins长度的毫秒值

        def fb_compare(ms: int, span: TimeSpan) -> int:
            assert ms >= 0 and span.start >= 0 and span.end > 0
            if ms < span.start:
                return -1
            elif span.start <= ms < span.end:
                return 0
            else:
                return 1

        def fa_iter(mss: Iterable[int]) -> Iterator[TimeSpan]:
            a, b = min(mss), max(mss)
            assert mss and 0 <= a <= b and L > 0
            start = (a // L) * L
            while True:
                span = TimeSpan(start=start, end=start + L)
                if fb_compare(b, span) >= 0:
                    yield span
                    start += L
                else:
                    return

        def fc_iter(span: TimeSpan, pms: int) -> Iterator[TimeSpan]:
            assert span.start >= 0 and span.end > 0 and pms >= 0, f'err70652 {span} {pms}'
            if fb_compare(pms, span) <= 0:
                return
            assert pms >= span.end
            for _s in fa_iter([span.end, pms]):
                if fb_compare(pms, _s) > 0:
                    yield _s
                else:
                    return

        def fd_rate(span: TimeSpan, mss: Iterable[DirSizeInfo]) -> int:
            """
            fd 两个参数 区间r和时间点集ps 计算区间内的网速数值
                ps是时间升序排列的
                sum(diff for a,b in pairwise(ps)
                    if b in span_r
                    if (diff := b.sizen -a.sizen)>0
                    )
            """
            s = sum(diff for a, b in ite.pairwise(mss)
                    if fb_compare(b.ctime_ms, span) == 0
                    if (diff := b.size_n - a.size_n) > 0
                    )
            return int(s / (TS_LEN / 1000))

        dus = self.dir_size_infos
        rts = self.rate_timespans

        def fcheck() -> None:
            AT.assert_(0 <= len(dus) < 100)
            # rts 也是严格升序的 而且是连续的 如果 len(rts)>100 直接pop(0)即可
            AT.assert_(all((b.start - a.start) == L for a, b in ite.pairwise(rts)), 'err39945')
            while len(rts) >= 100:
                rts.pop(0)
            AT.assert_(0 <= len(rts) < 100)
            AT.assert_(all(ts.rate >= 0 for ts in rts), 'err67961')

        fcheck()

        now = AT.fepochMillis()

        if not rts:
            # 最初始的状态
            sps = list(fa_iter([now]))
            AT.assert_(len(sps) == 1 and fb_compare(now, sps[0]) == 0, 'err54036')
            pre_span = sps[0].next_span(n=-1, rate=0)
            rts.append(pre_span.next_span(n=0, rate=fd_rate(pre_span, dus)))
            iprint_debug(f'add span network rates. {rts}')

        else:
            sps = [_span.next_span(n=0, rate=fd_rate(_span, dus)) for _span in fc_iter(rts[-1], now)]
            rts.extend(sps)
            if sps:
                iprint_debug(f'add span network rates. {sps}')
            fcheck()





from typing import cast

from dknovautils import *

from takahom.common.common_imports import run_simple_a
from takahom.common.common_util import CommonUtil, ChineseCvt
from takahom.common.common_etc import DkFile2, AT2, limit_memory

from attr import define, field as afield, asdict


def _f_rate_limits() -> str:
    a = [1000 * 1000 * 5] * 8 + [1000 * 300] * 16
    r = ','.join(str(e) for e in a)
    LMT = 1000 * 1000 * 10
    return f'*:{LMT}'


def trim_to_empty(v: str) -> str:
    r = v.strip()
    return r


@define(kw_only=True, frozen=True)
class ServerConfig:
    work_dir: str = afield(converter=trim_to_empty)
    profile: str = afield(default='')

    port: int = afield(default=0, )  # port>=0 如果是0 表示不启动web server

    secret: str = afield(default='', converter=trim_to_empty)

    # ---------------------------------------------
    # 暂时不手动设置

    scan_dirs: str = afield(default="", converter=trim_to_empty)

    network_proxy: str = afield(default="", converter=trim_to_empty)  # ='socks5://192.168.0.34:7890'

    rate_limits: str = afield(factory=_f_rate_limits, converter=trim_to_empty)

    chinese_cvt: int = afield(default=ChineseCvt.nothing.value)

    # ------------------------------------

    file_key: str = afield(default="TAK")

    p7z_volume: int = 1024 * 1024 * 10

    p7z_level: str = "-mx1"

    cfg_fid: int = 10 ** 7

    cfg_rid: int = 10 ** 5

    url_duplicate_check_days: int = 30

    # 网速检测 低于该网速则主动停止传输
    v130_limit_break: int = 1000 * 5

    ytb_cmd_path: str = "yt-dlp "

    s7z_keep_secs: int = 60 * 60

    state_res_max_size: int = 5

    # -----------------------------------------------------------
    conf_file: str = afield(default="")

    @work_dir.validator
    def check_workdir(self, attribute: Any, value: str) -> None:
        # 必须是目录 必须存在
        ma = f"work_dir:'{value}' doesn't exists or is not a directory "
        _dummy = value == ''
        if _dummy:
            return
        dkf = DkFile(value)
        if not dkf.exists() or not dkf.is_dir():
            raise ValueError(ma)

        tfile = DkFile2.join_path(dkf, f'tmp{AT.fepochMillis()}.tmp')
        try:
            with open(tfile.path, 'w') as file:
                file.write('hello')
        except Exception as e:
            raise ValueError(f"work_dir:'{value}' has no write permission ")
        finally:
            if tfile.exists():
                os.remove(tfile.path)

    #     secret
    @secret.validator
    def check_secret(self, attribute: Any, value: str) -> None:
        r = re.compile('[0-9a-zA-Z_]{6,30}')
        ok = value == '' or r.fullmatch(value)
        if not ok:
            sc = '*' * len(value)
            raise ValueError(f"secret:'{sc}' is bad. It should only contains 0-9a-zA-Z_ and 6<= len <=30 ")

    @network_proxy.validator
    def check_socksproxy(self, attribute: Any, value: str) -> None:
        r = re.compile('socks5://.*')
        ok = value == '' or r.fullmatch(value)
        if not ok:
            raise ValueError(f"socks_proxy:'{value}' is bad. It should starts with 'socks5://' ")

    @rate_limits.validator
    def check_ratelimits(self, attribute: Any, value: str) -> None:
        self.cal_rate_limits(value)

    # -------------------------------------------------------------------
    def cal_rate_limits(self, value: str) -> List[int]:
        rs = CommonUtil.parse_rate_limits_from_conf(value)
        assert len(rs) == 24
        return rs

    def cal_scan_dirs(self) -> List[DkFile]:
        rs = [DkFile(e) for e in self.scan_dirs.split(',') if e]
        return rs

    def copy_without_secret(self) -> Dict[str, Any]:
        osdc: Dict[str, Any] = asdict(self, filter=lambda attr, value: attr.name != "secret")
        return osdc


def test001() -> None:
    AT.logger.debug("hello from here")

    s = AT.md5_hex("".encode())
    assert len(s) > 0
    iprint(s)
    pass


def build_config_from_args(args: Any) -> ServerConfig:
    from dotenv import dotenv_values

    conf_file: str = args.conf
    conf_file = str(DkFile(conf_file).path.resolve())

    AT.assert_((dkf := DkFile(conf_file)).exists() and dkf.is_file(), f'err62869 file {conf_file} cannot be read')
    iprint_debug(f"conf file is '{conf_file}' ")

    cfg: Dict[str, str] = cast(Dict[str, str],
                               dotenv_values(conf_file))
    # config = {"USER": "foo", "EMAIL": "foo@example.org"}

    AT.assert_(all(v is not None for v in cfg.values()),
               f'err12922 some value is bad. check format of file {conf_file}')

    if cfg.get('work_dir', '').strip() == '':
        raise ValueError(f"err36901 work_dir is bad")

    _sc = ServerConfig(work_dir='')

    sc = ServerConfig(
        work_dir=cast(str, cfg.get('work_dir')),
        profile=args.profile,
        conf_file=conf_file,
        secret=cfg.get('secret', _sc.secret),
        port=int(cfg.get('port', _sc.port)),
        rate_limits=cfg.get('rate_limits', _sc.rate_limits),
        network_proxy=cfg.get('network_proxy', _sc.network_proxy),
        scan_dirs=cfg.get('scan_dirs', _sc.scan_dirs),
        chinese_cvt=int(cfg.get('chinese_cvt', _sc.chinese_cvt)),

    )

    CommonUtil.parse_rate_limits_from_conf(sc.rate_limits)

    return sc


"""



"""



from __future__ import annotations

from .common_imports import *
from .common_etc import AT2, DkFile2


def ln_filter_nor_comments_nor_empty(lns: Iterable[str]) -> Iterator[str]:
    """     每行都strip     去掉空行 #行 -行

    with "#", ";" or "]" are considered as
    comments and ignored

    """
    it = (ln for ln in lns if ln)
    it = AT2.ln_filter_not_startswith(it, ('#', ';', ']'))
    return it


def ln_filter_http_or_https(lns: Iterable[str]) -> Iterator[str]:
    return AT2.ln_filter_startswith(lns, ('http://', 'https://'))


class ChineseCvt(Enum):
    """
    pip install zhconv

    zh-cn 大陆简体
    zh-tw 台灣正體
    zh-hk 香港繁體


    """
    nothing = 0
    # 大陆简体
    zh_cn = 1
    # 台灣正體
    zh_tw = 2
    # 香港繁體
    zh_hk = 3


class CommonUtil:
    MIME_Stream = "application/octet-stream"
    MIME_text_html = "text/html"

    Err_Timeout = "err22603_Msg_timeout"

    special_check_urls_special_100 = [
        "https://totalcommander.ch/1102/tcmd1102x32.exe",
        "https://www.7-zip.org/a/7z2301-x64.exe",
        # 'https://github.com/git-for-windows/git/releases/download/v2.43.0.windows.1/Git-2.43.0-32-bit.exe',
        # 'https://github.com/git-for-windows/git/releases/download/v2.43.0.windows.1/Git-2.43.0-64-bit.exe',
        # 'https://github.com/git-for-windows/git/releases/download/v2.43.0.windows.1/PortableGit-2.43.0-32-bit.7z.exe',
    ]

    special_check_urls_ytb = [
        "https://www.youtube.com/watch?v=ntYmrW6c7GM",
    ]

    special_check_urls = special_check_urls_special_100 + []

    @staticmethod
    def clear_dir(tmp_dir: DkFile) -> None:
        DkFile2.clear_dir(tmp_dir)

    @staticmethod
    def rate_limit(lmts: List[int]) -> int:
        AT.assert_(
            len(lmts) == 24 and all(lm >= 1000 for lm in lmts),
            "err43170 bad rate_limits",
        )
        h = datetime.now().hour
        AT.assert_(0 <= h <= 23)
        return lmts[h]

    @staticmethod
    def test_random_github_url() -> str:
        return (
                CommonUtil.special_check_urls_special_100[0]
                + f"?r={random.randint(10 ** 4, 10 ** 5)}"
        )

    @staticmethod
    def parse_resolution_of_input_filename(dkf: DkFile) -> int | None:
        """
        输入文件的匹配规则
            文件名示例 ytb_xxxx_0720.urls.txt

            文件名不包括 copy 字符串 (因为某些编辑器会临时自动生成这种文件名)
            文件名 ytb_ 开头 .urls.txt结尾
            去掉扩展名 下划线_划分为n部分 n>=2
            s[-1]计算分辨率

            如果只有2部分 则 s[1]计算分辨率

        """
        fname = dkf.basename

        if "copy" not in fname and re.fullmatch(reg_ytb_url_txt, fname):
            fns = fname.split(".")[0].split("_")
            AT.assert_(len(fns) >= 2, f"err50751 bad filename {dkf.pathstr}")
            r = int(fns[-1])
            AT.assert_(240 <= r <= 2500, f"err87022 bad res {dkf.pathstr}")
            return r
        else:
            return None

    @staticmethod
    def url_simple_meta(url: str) -> str:
        # urlinfo = DLCommon.url_simple_meta(url) # 0-9a-zA-Z
        ptn = r"[^0-9a-zA-Z]+"
        url = re.sub(ptn, "-", url)[:70]
        return url

    @staticmethod
    def parse_urls_from_sinput_text(txt: str) -> Iterator[str]:
        return CommonUtil.parse_urls_from_txt(txt)

    @staticmethod
    def parse_urls_from_txt(txt: str) -> Iterator[str]:
        """
        每行去掉头尾空白
        移除 # - 开头的行
        行内 用空白符分割，取0索引部分
        该部分用 urlparser解析，获取url正常 则作为正常url返回

        """
        from urllib.parse import urlparse

        lns = txt.strip().splitlines()
        lns = ln_filter_nor_comments_nor_empty(lns)
        lns = (ln.split(" ")[0] for ln in lns)
        d = {ln: None for ln in lns}
        lns = list(d.keys())

        for ln in lns:
            rp = urlparse(ln)
            if rp.scheme and rp.netloc:
                yield ln

    @staticmethod
    def dir_has_filename(d: DkFile, fname: str) -> bool:
        p = join(d.pathstr, fname)
        p = DkFile(p)
        return p.exists() and p.is_file()

    @staticmethod
    def parse_filename_dict(df: DkFile | str, check: bool = False, silent: bool = False) -> Dict[str, str]:
        """
        如果发生格式错误 将返回空集

        """
        d: Dict[str, str] = {}
        try:
            basename = df.basename if isinstance(df, DkFile) else df.strip()
            bname = basename.split(".")[0]
            bks = (x for x in bname.split("_") if x)  # 防止头部和尾部的_符号造成的空字符串
            for kv in bks:
                ks = kv.split("-")
                _ok = len(ks) >= 2 and len(ks[0]) > 0
                if not _ok:
                    if silent:
                        return {}
                    else:
                        assert False, f"err82633 {ks} {df}"
                k = ks[0]
                v = kv[len(k) + 1:]
                d[k] = v

            if check:
                dc = d
                dcok = dc["ln"] and dc["md5"] and dc["epoch"]
                AT.assert_(dcok, "err48065")

            return d
        except Exception as e:
            if silent:
                iprint_debug(e)
                return {}
            else:
                raise e

    @staticmethod
    def parse_rate_limits_from_conf(s: str) -> List[int]:
        return CommonUtil.parse_rate_limits_from_conf3(s)

    @staticmethod
    def parse_rate_limits_from_conf3(rates_conf: str) -> List[int]:
        """

网速的设置方法
    用逗号分成segment 星号匹配所有小时 0-3表示0,1,2,3一共4个小时的区间
    不同的segment 如果重复定义 以前面的优先级最高
    8:23424,0-3:32424,4-6:234324234,*:3242424

        """

        def pa(s: str) -> Dict[int, int]:

            def mint(s: str) -> int:
                s = int(s.strip())
                assert 0 <= s < 24, f'err91030 bad hour {s}'
                return s

            s = s.strip()
            hp, rate = s.split(':')
            hp = hp.strip()
            rate = int(rate)
            assert 3000 <= rate, f'err53866 need rate >= 3000. bad rate {rate}'

            hs = []
            if hp == '*':
                hs = list(range(0, 24))
            elif '-' not in hp:
                hs = [mint(hp)]
            elif '-' in hp:
                ha, hb = [mint(x) for x in hp.split('-')]
                assert ha <= hb, f'err34531 bad hour {hp}'
                hs = list(range(ha, hb + 1))
            else:
                assert False, f'err89451 bad hour {hp}'

            return {h: rate for h in hs}

        rates_conf = rates_conf.strip()
        LMT = 1000 * 1000 * 10
        dc = {k: LMT for k in range(0, 24)}
        dcs = [pa(p2) for p in rates_conf.split(',') if (p2 := p.strip())]
        for d in reversed(dcs):
            dc.update(d)
        return list(dc.values())

    @staticmethod
    def chinese_cvt_str(fname: str, cvt: ChineseCvt) -> str:
        import zhconv  # type:ignore
        if cvt == ChineseCvt.nothing:
            return fname
        mdict = {
            ChineseCvt.zh_cn: 'zh-cn',
            ChineseCvt.zh_tw: 'zh-tw',
            ChineseCvt.zh_hk: 'zh-hk',
        }

        assert len(mdict) + 1 == len(list(ChineseCvt))
        r = cast(str, zhconv.convert(fname, mdict[cvt]))
        return r

    @staticmethod
    def chinese_convert_directory(pdir: DkFile, cvt: ChineseCvt) -> None:
        """
        将pdir目录下（包括该目录）的所有文件名，包括目录名，都进行重命名
        buttoneup的方式遍历，用rename

import zhconv
text = '便纵有千种风情，更与何人说？'
text1 = zhconv.convert(text, 'zh-hant')

        """

        cnt = 0

        def fcvt(dkf: DkFile, cvt: ChineseCvt) -> None:
            h, t = os.path.split(dkf.path)
            bn = dkf.basename
            bn2 = CommonUtil.chinese_cvt_str(bn, cvt)
            if bn != bn2 and not DkFile(join(h, bn2)).exists():
                os.rename(dkf.pathstr, join(h, bn2))
                nonlocal cnt
                cnt += 1

        for dkf in DkFile2.oswalk_simple_a(pdir, topdown=False):
            fcvt(dkf, cvt)

        if cnt:
            iprint_info(f'convert chinese {cnt}')

    @staticmethod
    def _fcvt(dkf: DkFile, fname: Callable[[DkFile], str]) -> DkFile | None:
        h, t = os.path.split(dkf.path)
        bn = dkf.basename
        bn2 = fname(dkf)
        ndkf = DkFile(join(h, bn2))
        if bn != bn2 and not ndkf.exists():
            os.rename(dkf.pathstr, join(h, bn2))
            return ndkf
        return None

    @staticmethod
    def dir_replace_filename_comma(pdir: DkFile) -> None:
        """
        将所有文件名中的逗号替换为-
        """
        for dkf in DkFile2.oswalk_simple_a(pdir):
            CommonUtil._fcvt(dkf, lambda dkf: dkf.basename.replace(',', '-'))



from __future__ import annotations

from takahom.common.down_request import DownRequest
from takahom.common.common_etc import limit_memory, AT2
from takahom.server.server_base_a import KeyEvent
from takahom.server.server_base_web import IServerBaseWeb
from takahom.server.server_base_e import INetCalulator
from takahom.server.server_index import DLServerIndex

from takahom.server.server_ytb_util import *


class IServerBaseC(IServerBaseWeb, INetCalulator, Protocol):
    def thread_main_download(self, cnt: int) -> None:
        """

        扫描全部 sinput/*.reqa.toml 文件
            将ctt适当预处理
            获取1条任务 排除已经执行的任务

                每个 sdownload/r-1312313/ 目录下 有文件 info.json 其中有对应的url

        调用下载命令，获取返回结果


        下载失败 记录状态

        下载成功
            目录内的数据进行移动 sdownload/r-123/下面
                创建 info.json文件

            在s7z目录下创建 7z文件
                s7z/
                r-123.7z.1
                r-123.7z.2

            在 sresp目录下创建 data 文件
                fid-123_r-123_n-3_i-1_md5-xxx.data

        下载基本结果，保存到index文件中
            从内存更新到index文件 用json格式 不然代码升级可能导致数据不能读取

        """

        dir_scache = self.subdir("scache")
        dir_sinput = self.subdir("sinput")
        dir_sdownload = self.subdir("sdownload")
        dir_s7z = self.subdir("s7z")
        dir_sresp = self.subdir("sresp")
        dir_stmp = self.subdir("stmp")

        req: DownRequest | None = None

        req = self.fetch_req()
        assert req

        iprint_info(f"start download: {req.url}")

        res = self.handle_request(req)

        res.last_new_urls = self.last_new_urls
        assert self.state_index
        if vAddRes := False:
            rss = self.state_index.results
            rss.append(res)
            while len(rss) > self.sconf.state_res_max_size:
                rss.pop(0)

        if res.ok or not res.ok:
            rid = self.state_index.rid
            res.rid = rid

            # 尽早持久化 防止rid重复
            self.state_index.rid += 1
            self.write_index_file()

            res.write_json_file()

            # scache中的所有文件名 确保将,替换为-
            CommonUtil.dir_replace_filename_comma(dir_scache)

            iprint_info("will copy to sdownload dir 10secs later")
            time.sleep(10)

            title_part = YtbUtil.cal_title_part_from_description_filename(res, dir_scache)
            if title_part:
                title_part = CommonUtil.chinese_cvt_str(title_part, cvt=self.chinese_cvt_enum)
            iprint_debug(f'step73261 {title_part}')
            sdownload_dir_name = res.cal_sdownload_dir_name(ctitle=title_part)
            sdownload_dir_path = join(dir_sdownload.pathstr, sdownload_dir_name)
            os.makedirs(sdownload_dir_path)
            AT.assert_((_ := DkFile(sdownload_dir_path)).is_dir() and _.exists(), "err38182")
            shutil.copytree(req.data_dir, sdownload_dir_path, dirs_exist_ok=True)
            # 目录下所有文件都简体和繁体转换
            CommonUtil.chinese_convert_directory(DkFile(sdownload_dir_path), cvt=self.chinese_cvt_enum)

            iprint_info(
                f"end download. ok:{res.ok} url:{req.url} network: {int(res.netv / 1000)}KB/s"
            )

        if not res.ok:
            self.write_index_file()
            iprint_info(f"err72093 donwlaod error {res.to_str_simple()}")
            self.req_running = None
        else:
            """
            目录内的数据进行移动 sdownload/下面
                创建 info.json文件

            在s7z目录下创建 7z文件
                s7z/
                r-123.7z.001
                r-123.7z.002

            在 sresp目录下创建 data 文件
                fid-123_r-123_n-3_i-1_md5-xxx.data

            """
            iprint_debug("step54059")
            # fid = self.state_index.fid
            fid = 10000
            AT.assert_(res.files_cnt > 0 and res.files_size > 0, "err79860")

            run_simple_a(f'ls -l "{sdownload_dir_path}" ')

            # 7za a -p123456 -mhe -mx0 -t7z -v3m directory.7z directory
            # 压缩的文件中是有目录名称的
            pwd = self.secret
            volume = 1024 * 1024
            volume = self.p7z_volume

            def f_vlm(fsize: int, vlm: int) -> int:
                AT.assert_(fsize > 0 and vlm >= 1000, "err93746")
                while True:
                    fnum = int(fsize / vlm) + 1
                    if fnum < 300:
                        return vlm
                    else:
                        vlm = int(vlm * 1.2)

            volume = f_vlm(res.files_size, volume)
            # AT.assert_(res.files_size/volume<900,'err67222 too many volumes')
            AT.assert_(req.ytb_ytbid, "err32546")
            # bname_7z = f"r-{self.rid}_{req.ytb_ytbid.to_str_with_prefix()}.{self.fkey}.7z"
            bname_7z = res.cal_s7z_partial_name(p7z_vn=0, secret=self.secret, fkey=self.fkey)
            run_simple_a(f'mkdir "{dir_s7z.pathstr}"/tmp ')
            run_simple_a(f'rm -rf "{dir_s7z.pathstr}"/tmp/* ')
            dest_s7z_file_path = f"{dir_s7z.pathstr}/tmp/{bname_7z}"
            mxn = self.p7z_level
            pwd = pwd.strip()
            _ppwd = f'-p{pwd}' if pwd else ''
            _cmd = (
                f'7za a {_ppwd} -mhe {mxn} -t7z -v{volume}b "{dest_s7z_file_path}" "{sdownload_dir_path}" '
            )
            run_simple_a(_cmd)
            iprint_debug("step30466")
            # 可能文件size太小 那么就没有分卷 为了统一处理 添加后缀 001
            if DkFile(dest_s7z_file_path).exists():
                run_simple_a(f'mv "{dest_s7z_file_path}" "{dest_s7z_file_path}.001" ')

            # 验证 rid对应的7z文件 创建成功 ls文件
            fs = [
                dkf
                for dkf in DkFile.listdir(join(dir_s7z.pathstr, 'tmp'))
                if bname_7z in dkf.basename
            ]
            AT.assert_(len(fs) > 0, "err21761")
            iprint_debug(f'step81273 fs {",".join([dkf.basename for dkf in fs[:100]])}')

            # 计算创建的7z的文件个数
            p7z_vn = len(fs)
            AT.assert_(0 < p7z_vn < 900, "err23868")

            # 将文件从tmp mv up 同时重命名
            for i in range(1, p7z_vn + 1):
                suffix = str(1000 + i)[1:]
                from_path = f'{dir_s7z.pathstr}/tmp/{bname_7z}.{suffix}'
                bname_7z_new = res.replace_s7z_vn(bname_7z, p7z_vn=p7z_vn)
                to_path = f'{dir_s7z.pathstr}/{bname_7z_new}.{suffix}'
                run_simple_a(f'mv "{from_path}" "{to_path}" ')

            # 尽早持久化 防止fid重复
            # self.state_index.fid += n
            self.write_index_file()

            self.req_running = None
            res.infos.append("step100 save file ok")
            CommonUtil.clear_dir(dir_scache)
            self.log_cnt(KeyEvent.down_allok)

            iprint_debug("step65252")



from __future__ import annotations

from enum import auto
from typing import Final

from takahom.common.down_request import DownRequest
from takahom.common.common_etc import limit_memory, AT2
from takahom.server.server_config import ServerConfig
from takahom.server.server_index import DLServerIndex

from takahom.server.server_ytb_util import *


class KeyEvent(Enum):
    down_begin = auto()
    down_good = auto()
    down_bad = auto()
    down_error = auto()
    down_end = auto()
    down_allok = auto()
    down_end_demo = auto()

    file_begin = auto()
    file_ok = auto()
    file_error = auto()


Source_Demo = 'demo'


class IServerBaseA(Protocol):
    # readable and writable property
    args: Any
    work_dir: DkFile
    state_index: DLServerIndex
    sconf: ServerConfig
    start_app_epoch_str: str
    port: int
    secret: str
    fkey: str
    p7z_volume: int
    p7z_level: str
    s7z_keep_secs: int
    ytb_cmd_path: str
    socks_proxy: str
    scan_dirs: List[DkFile]
    input_dir: DkFile
    output_dir: DkFile

    req_running: DownRequest | None
    downloading_url: str | None
    downloading_begin: str | None
    scache_size: int

    rate_limits: List[int]
    last_new_urls: int
    downloaded_bytes: int

    req_queue: Queue[DownRequest]
    req_queue_list: List[YtbItem]
    req_size_max: int

    stt_cnts: Dict[str, int]

    def fetch_req(self) -> DownRequest:
        req = self.req_queue.get()
        self.req_queue_list.remove(req.ytb_item)
        return req

    def put_req(self, req: DownRequest) -> None:
        AT.assert_(req.check(), "err59453")
        self.req_queue.put(req)
        self.req_queue_list.append(req.ytb_item)

    def subdir(self, d: str) -> DkFile:
        r = DkFile(join(self.work_dir.path, d))
        AT.assert_(r.is_dir() and r.exists(), f"err57249 {r.path}")
        return r

    def get_index_file(self) -> DkFile:
        idxfile = DkFile(join(self.work_dir.pathstr, "index.json"))
        return idxfile

    def read_index_file(self) -> DLServerIndex | None:
        sfile = self.get_index_file()
        if sfile.exists():
            s = sfile.path.read_text(encoding="utf-8")
            r: DLServerIndex = DLServerIndex.from_json(s)  # type:ignore
            return r
        else:
            return None

    @property
    def chinese_cvt_enum(self) -> ChineseCvt:
        return ChineseCvt(self.sconf.chinese_cvt)

    def f_handle_test(self, req: DownRequest) -> DownResult:
        # 在目录下写入随机文件即可
        start_epoch = AT.fepochSecs()
        resid = random.randint(10 ** 4, 10 ** 4 * 2)
        demostr = "".join(str(i) for i in range(resid))

        demostr = req.to_json()  # type:ignore
        Path(join(req.data_dir, "TEST.json")).write_text(demostr)
        time.sleep(5)
        end_epoch = AT.fepochSecs()
        dr = DownResult(
            req=req,
            ok=True,
            data_dir=req.data_dir,
            url=req.url,
            res_id=str(resid),
            infos=[],
            start_epoch=start_epoch,
            end_epoch=end_epoch,
        )

        dr.after_run()
        return dr

    def f_handle_simple_curl_100(self, req: DownRequest) -> DownResult:
        start_epoch = AT.fepochSecs()
        resid = random.randint(10 ** 4, 10 ** 4 * 2)
        demostr = "".join(str(i) for i in range(resid))

        url = req.url

        # AT.assert_(url.split('?'))

        u2 = url.split("?")[0]

        fname = u2.split("/")[-1]

        cmd = f"curl '{url}' --output '{req.data_dir}/{fname}' "
        rc, out, err = run_simple_b(cmd)

        assert isinstance(err, str)

        end_epoch = AT.fepochSecs()

        dr = DownResult(
            req=req,
            ok=(rc == 0),
            data_dir=req.data_dir,
            url=req.url,
            res_id=str(resid),
            infos=[],
            err=err,
            start_epoch=start_epoch,
            end_epoch=end_epoch,
        )

        dr.after_run()

        dr.infos.append(f"cmd: {cmd}")
        return dr

    def f_dev_100(self) -> None:
        dir_scache = self.subdir("scache")

        req = DownRequest(
            url=CommonUtil.test_random_github_url(),
            data_dir=dir_scache.pathstr,
            uuid=AT2.gen_uuid_str(),
        )
        req.domain = "github"
        req.resource_key = str(hash(req.url))

        req.init()
        res = self.f_handle_simple_curl_100(req)
        AT.assert_(
            res.files_cnt > 0 and res.files_size > 0,
            f"err57447 {res.to_json()}",  # type:ignore
        )

    def log_cnt(self, key: Any, v: int = 1, level: LLevel = LLevel.Debug) -> None:

        match key:
            case KeyEvent():
                k = key.name
            case _:
                k = str(key)

        self.stt_cnts[k] = self.stt_cnts.get(k, 0) + v
        iprint(f'log cnt {k} {v}', level)

    def write_index_file(self) -> None:
        sstate = self.state_index
        s = sstate.to_json(indent=4, ensure_ascii=False)  # type:ignore
        self.get_index_file().path.write_text(s, "utf-8")
        return

    def handle_request(self, req: DownRequest) -> DownResult:

        self.log_cnt(KeyEvent.down_begin)
        iprint_info(f'download url begin: {req.url}')

        r: DownResult | None = None
        try:
            dir_scache = self.subdir("scache")

            self.req_running = req
            self.downloading_url = req.url
            self.downloading_begin = AT.sdf_logger_format_datetime()

            if True:
                req.data_dir = dir_scache.pathstr
                req.start_epoch = AT.fepochMillis()
                req.rate_limit = CommonUtil.rate_limit(self.rate_limits)
                CommonUtil.clear_dir(dir_scache)

            AT.assert_(req.url and req.data_dir, "err40069")
            url = req.url

            if False:
                AT.never()
            elif req.url.startswith("http://test/"):
                return self.f_handle_test(req)
            elif (url.startswith("https://github.com/")
                  or url.startswith("https://totalcommander.ch/")
                  or f_url_in_special_100(url)
                  or False):
                return self.f_handle_simple_curl_100(req)
            elif YtbUtil.parse_ytbitem_from_url(url):
                r = ytb_dl_run(req)
                AT.assert_(r, "err35277")
                assert r

                AT.assert_(r.start_epoch > 0 and r.end_epoch > 0, "err19241")
                r.after_run()

                return r

            else:
                AT.never()
                assert False
        except Exception as e:
            self.log_cnt(KeyEvent.down_error, level=LLevel.Warn)
            raise e
        else:
            pass

        finally:
            self.log_cnt(KeyEvent.down_end)
            iprint_info(f'download url end: {req.url}')
            self.req_running = None
            self.downloading_url = None
            self.downloading_begin = None

            if r and r.req.source == Source_Demo:
                self.log_cnt(KeyEvent.down_end_demo)
            if r and r.ok:
                self.log_cnt(KeyEvent.down_good)
            elif r and not r.ok:
                self.log_cnt(KeyEvent.down_bad)
            else:
                pass


def f_url_in_special_100(url: str) -> bool:
    return any(
        url.startswith(u) for u in CommonUtil.special_check_urls_special_100
    )


def dec_my_log_error_builder(message: str) -> Any:
    # 虽然已经全部函数参数都赋值 但类型仍然是 dec_builder 所以必须eval之后才能返回
    dec_log_on_err_builder = functools.partial(
        log_on_error,
        log_level=logging.ERROR,
        message=message,
        logger=AT.logger,
        on_exceptions=(IOError, Exception),
        reraise=True,
    )
    return dec_log_on_err_builder()


def copy_ytbtxts_files_to_dest_input_dir(src: DkFile, dest: DkFile) -> None:
    """
    如果目标目录中没有任何文件
    则将源目录中存在 ytb_*.txt 文件 全部拷贝过去

    """

    copy = dest.exists() and dest.is_dir() and not list(DkFile2.oswalk_simple_a(dest)) \
           and src.exists() and src.is_dir()
    if not copy:
        return

    for dkf in DkFile2.oswalk_simple_a(src, only_file=True):
        if re.fullmatch(ytb_reg_ytb_txt_files, dkf.basename):
            shutil.copyfile(dkf.path, DkFile2.join_path(dest, dkf.basename).path)


# @dataclass_json
@define
class SystemInfo:
    uname: str
    memory: str
    ffmpeg: str
    ytdlp: str
    p7zip: str
    wget: str
    work_du: str


def build_system_info(work_dir: DkFile) -> SystemInfo:
    """

    """
    AT.assert_(AT.is_linux, 'err32791 only support linux')

    _, uname, _ = run_simple_b('uname -a')
    _, memory, _ = run_simple_b('free -mh |head -n 5')
    _, ffmpeg, err = run_simple_b('ffmpeg --version 2>&1 |head -n 5')
    """
ffmpeg version 4.2.7-0ubuntu0.1 Copyright (c) 2000-2022 the FFmpeg developers
  built with gcc 9 (Ubuntu 9.4.0-1ubuntu1~20.04.1)    
    """
    AT.assert_('FFmpeg' in ffmpeg and 'Copyright' in ffmpeg,
               f'err87868 ffmpeg cannot run. {ffmpeg} {err}')

    # yt-dlp --version
    _, ytdlp, err = run_simple_b(f'yt-dlp --version')
    # return: 2023.12.30
    AT.assert_(len(ytdlp) > 4, f'err15694 yt-dlp cannot run. {err}')

    _, p7zip, err = run_simple_b('7za -h|head -n 5')
    """

7-Zip (a) [64] 16.02 : Copyright (c) 1999-2016 Igor Pavlov : 2016-05-21
p7zip Version 16.02 (locale=C.UTF-8,Utf16=on,HugeFiles=on,64 bits,12 CPUs Intel(R) Core(TM) i7-10750H CPU @ 2.60GHz (A0652),ASM,AES-NI)

Usage: 7za <command> [<switches>...] <archive_name> [<file_names>...]

    """
    AT.assert_('Copyright' in p7zip, f'err53910 7za cannot run. need to install p7zip. {err}')

    # 也可以用 xxx -h 2>/dev/null | head -n 1 但是这样就没有版本信息
    cmd = 'wget --version | head -n 5'
    _, wget, _ = run_simple_b(cmd)
    """
GNU Wget 1.20.3 built on linux-gnu.

-cares +digest -gpgme +https +ipv6 +iri +large-file -metalink +nls
+ntlm +opie +psl +ssl/openssl
    """
    AT.assert_('Wget' in wget, f'err21171 wget cannot run')

    # 重要目录的磁盘占用
    AT.assert_(work_dir.exists() and work_dir.is_dir(), f'err40659 dir is bad {work_dir}')

    # du . -d1 -h -BG |sort -r
    cmd = f'du "{work_dir.pathstr}" -d1 -h -BG |sort -r'
    _, work_du, _ = run_simple_b(cmd)
    work_du = ' '.join(work_du.splitlines())

    info = SystemInfo(
        uname=uname.strip(),
        memory=memory.strip(),
        ffmpeg=ffmpeg.strip(),
        ytdlp=ytdlp.strip(),
        p7zip=p7zip.strip(),
        wget=wget.strip(),
        work_du=work_du.strip(),
    )

    return info



try:
    import sys

    print(f'sys.path: {sys.path}')
finally:
    pass

from takahom.server import server

if __name__ == "__main__":
    def f_parse_args() -> None:
        import argparse
        parser = argparse.ArgumentParser()

        parser.add_argument("mode", choices=["demo", "run"], help="subcmd")

        parser.add_argument("--profile", type=str,
                            default="prod", choices=["prod", "dev"], help="profile")

        parser.add_argument("--conf", type=str, default='env.conf', help="config file")

        args = parser.parse_args()

        server.main(args)


    f_parse_args()



from __future__ import annotations

from takahom.server.server_ytb_util import *


class _DemoAppCtx(object):
    def __init__(self, args: Any) -> None:
        self.rate_limits: List[int] = []
        self.socks_proxy: str = ""  # ='socks5://192.168.0.34:7890'
        self.cmd_path: str = ""
        self.work_dir = args.workdir
        self.scan_dirs: List[DkFile] = []
        self.input_dir = DkFile(join(self.work_dir, "input"))
        self.output_dir = DkFile(join(self.work_dir, "output"))
        self.tmp_dir = DkFile(join(self.work_dir, "tmp"))

        assert self.input_dir.exists(), self.input_dir.pathstr


appCtx: _DemoAppCtx | None = None


def f_download_simple_a(args: Any) -> None:
    """

    循环扫描目录

    ytb_2500.txt
    ytb_720.txt
    ytb_360.txt

    获取1个url 然后下载

    子目录名  r-100_ct-202312T123456_ytbid-312313_c-{ok|err}_url-{url简化的info}


    """
    assert appCtx, "err78385"
    scnt = 0
    while True:
        # 可以用文件让循环主动break
        if CommonUtil.dir_has_filename(appCtx.input_dir, "exit.cmd"):
            iprint_warn("step17239 file exit.cmd found. exit main loop.")
            break

        scnt += 1
        if scnt != 1:
            t = 30 + random.uniform(0, 30)
            iprint_debug(f"sleeping {t}s")
            time.sleep(t)

        # 排除 excludecodes， output_dir中的重复，选择一个url
        exids = set(ytb_scan_ytbids_excludes_file(appCtx.input_dir)) | set(
            ytb_scan_output_ytbids_scan_dirs(
                scan_dirs=appCtx.scan_dirs, output_dirs=[appCtx.output_dir]
            )
        )
        reqs = [
            req
            for req in ytb_parse_input_dir_reqs(
                input_dir=appCtx.input_dir,
                socks_proxy=appCtx.socks_proxy,
                # write_dir=appCtx.tmp_dir,
                cmd_path=appCtx.cmd_path,
            )
            if req.ytb_ytbid not in exids
        ]

        if not reqs:
            iprint_debug(f"no new url {scnt}")
            continue

        iprint_info(f"find new urls {len(reqs)}")
        req: DownRequest = reqs[0]

        iprint_info(f"start download: {req.url}")
        req.data_dir = appCtx.tmp_dir.pathstr
        req.start_epoch = AT.fepochSecs()
        req.rate_limit = CommonUtil.rate_limit(appCtx.rate_limits)

        CommonUtil.clear_dir(appCtx.tmp_dir)
        if not DkFile(req.data_dir).exists():
            os.mkdir(req.data_dir)
        res: DownResult | None = ytb_dl_v100(req)
        assert res
        AT.assert_(res.start_epoch > 0 and res.end_epoch > 0, "err19241")
        res.after_run()

        iprint_info(f"end download: {req.url} network: {int(res.netv / 1000)}KB/s")

        res.write_json_file()

        # 将目录 mv 到output 的子目录中 不会mv的写法 只好copy
        res.rid = 100
        rdirname = res.cal_sdownload_dir_name()
        shutil.copytree(req.data_dir, join(appCtx.output_dir.pathstr, rdirname))

        if res and res.ok:
            # 生成一个完整的7z文件 可以加上密码 定义在 DLCommon 类函数中 todo
            pass

        else:
            # 记录错误日志
            pass

        iprint_info(f"end job. url:{req.url} ok:{res.ok}")


def f_excludecodes20231228() -> str:
    # AT.unimplemented() ytbid-xxxxx
    return """



""".strip()


def _f_learn_dlp_100() -> None:
    import json
    import yt_dlp  # type:ignore

    URL = "https://www.youtube.com/watch?v=BaW_jenozKc"

    # ℹ️ See help(yt_dlp.YoutubeDL) for a list of available options and public functions
    ydl_opts: Dict[Any, Any] = {}
    with yt_dlp.YoutubeDL(ydl_opts) as ydl:
        info = ydl.extract_info(URL, download=False)

        # ℹ️ ydl.sanitize_info makes the info json-serializable
        print(json.dumps(ydl.sanitize_info(info)))

    pass


if __name__ == "__main__":
    import argparse  # 1

    parser = argparse.ArgumentParser()  # 2

    parser.add_argument("-p", "--profile", type=str, default="dev", help="profile")

    parser.add_argument("--location", type=str, default="test", help="location")

    parser.add_argument("--workdir", type=str, default=r"", help="工作目录")

    # parser.add_argument('--datadir',
    #                     type=str, default=r'q:\webdav_share\hesai_data\raw', help='数据目录')

    # 最多包括目录个数
    parser.add_argument("--maxcnt", type=int, default=50, help="dircnt")

    # 是否运行测试工况。可以用字符串详细描述 例如 1,2,3,4 等等
    parser.add_argument("--testcase", type=str, default="", help="是否运行测试工况")

    parser.add_argument("--reportverbose", type=int, default=30, help="报告的详细程度")

    args = parser.parse_args()  # 4

    if AT.is_windows:
        args.workdir = r"d:\tmp_dl"
    else:
        args.workdir = r"/mnt/ss/_DISC_S_/ytb_download"

    appCtx = _DemoAppCtx(args)

    appCtx.rate_limits = [1000 * 1000 * 3] * 8 + [1000 * 300] * 16

    if AT.is_windows:
        appCtx.rate_limits = [1000 * 300] * 24
    else:
        pass

    CommonUtil.rate_limit(appCtx.rate_limits)

    if AT.is_windows:
        appCtx.cmd_path = r"f:\DIKI\DIKIDEV\ZTools\yt-dlp2023.exe "
    else:
        appCtx.cmd_path = r"yt-dlp "

        appCtx.scan_dirs.append(DkFile("/mnt/ss/_DISC_S_/ytb_scan"))

    appCtx.socks_proxy = "socks5://192.168.0.34:7890"

    f_download_simple_a(args)


def _urlsa() -> str:
    return """


#girl show
https://www.youtube.com/watch?v=nFmmumBB1_4

#古筝 hotel california
https://www.youtube.com/watch?v=gf6v59c5yuY

https://www.youtube.com/watch?v=hgVq-DpCtao
https://www.youtube.com/watch?v=8syimRWomrs
https://www.youtube.com/watch?v=AMakl19CckA
https://www.youtube.com/watch?v=ynQypJe8k2g
https://www.youtube.com/watch?v=t1nlFJV5csM
https://www.youtube.com/watch?v=IOiedWxP5Zw
https://www.youtube.com/watch?v=n4WNxRbYwB4
https://www.youtube.com/watch?v=jqirCnIIKC4
https://www.youtube.com/watch?v=tSGD1AYrZg4&t=42s
https://www.youtube.com/watch?v=SKAkuZpMCPQ
https://www.youtube.com/watch?v=gf6v59c5yuY&list=RDgf6v59c5yuY&start_radio=1
https://www.youtube.com/watch?v=ighkK6pKJng


""".strip()


"""


"""



from __future__ import annotations

import re

from .server_imports import *
from takahom.common.down_request import DownRequest
from takahom.common.entity import YtbId, YtbItem
from takahom.common.down_result import DownResult
from takahom.common.common_etc import DkFile2
from takahom.common.ytb_util import YtbUtil, ln_filter_nor_comments_nor_empty


def ytb_dl_run(req: DownRequest) -> DownResult | None:
    return ytb_dl_v100(req=req)


def ytb_parse_input_dir_reqs(
        input_dir: DkFile, socks_proxy: str, cmd_path: str
) -> List[DownRequest]:
    """
    # 返回全部文件中的所有url 有顺序问题

    """

    dkfs = [
        dkf
        for dkf in DkFile2.oswalk_simple_a(input_dir, only_file=True)
        if CommonUtil.parse_resolution_of_input_filename(dkf)
    ]
    dkfs = sorted(dkfs, key=lambda dkf: dkf.basename)
    iprint_debug(f"find input files. {len(dkfs)} {[f.basename for f in dkfs]}")

    reqs: List[DownRequest] = []
    ytbids: Set[YtbId] = set()
    for dkf in dkfs:
        urls = CommonUtil.parse_urls_from_sinput_text(dkf.path.read_text())
        res = CommonUtil.parse_resolution_of_input_filename(dkf)
        assert res, "err98972"
        AT.assert_(240 <= res <= 2500, "err66968")
        for urla in urls:
            ytbitem = YtbUtil.parse_ytbitem_from_url(urla)
            if not ytbitem:
                iprint_info(f"err69431 bad url {urla} file:{dkf}")
                continue
            assert ytbitem
            req = DownRequest(url=ytbitem.get_url(), ytb_resolution=res)
            req.init()
            req.ytb_item = cast(YtbItem, ytbitem)
            req.ytb_ytbid = cast(YtbId, YtbUtil.parse_ytbid_from_url(ytbitem.get_url()))
            req.domain = ytb_domain_name
            req.resource_key = req.ytb_ytbid.to_str_with_prefix()
            req.proxy = socks_proxy
            req.cmd = cmd_path
            # req.source = f"file-{dkf.basename}"
            req.source = f"{dkf.basename}"

            if req.ytb_ytbid not in ytbids:
                ytbids.add(req.ytb_ytbid)
                reqs.append(req)

    return reqs


def ytb_scan_output_ytbids_scan_dirs(
        scan_dirs: List[DkFile], output_dirs: List[DkFile]
) -> Iterator[YtbId]:
    """
    深度扫描所有目录
    深度扫描所有文件

    """

    dkfs = (dkf.basename
            for dkdir in (scan_dirs + output_dirs)
            for dkf in DkFile2.oswalk_simple_a(dkdir))

    ids = YtbUtil.parse_ytbid_from_lines(dkfs)
    return ids


def ytb_scan_ytbids_excludes_file(input_dir: DkFile) -> Iterator[YtbId]:
    """

    """
    # ytb_*.exd.txt

    dkfs = (dkf for dkf in DkFile2.oswalk_simple_a(input_dir, only_file=True)
            if re.fullmatch(ytb_reg_exd_files, dkf.basename))

    for dkf in dkfs:
        txt = dkf.path.read_text("utf-8")
        codesa = YtbUtil.parse_ytbid_from_lines(txt.splitlines())
        for id in codesa:
            yield id


def ytb_dl_v100(req: DownRequest) -> DownResult | None:
    """


    偶尔存在 极低概率存在 子进程僵化的现象。没有任何反应 下载进程等待10个小时 没有任何下载 也不退出

    本应用中的cmd运行中的 stdout信息被捕获 在控制台或者日志中是不能实时查看的
        有时候运行cmd之后 目录内没有新的数据生成 进程也不返回 这时候应该是cmd出现问题了。
        此时的一个调试方法是 将cmd从日志中拷贝出来，将程序退出，然后单独运行该cmd 看out信息 可以了解具体的过程

        多线程的启动子进程

        while True:
            if 目录内所有普通文件的sum_filesize 超过 N=120s 没有任何变化:
                kill process
                sleep 10s
                确认线程正常退出
                break
            else:
                sleep 30s


    """
    # 说明：
    #     错误的url 必然返回1 而且有一定的原因说明
    #     网络中断 也能返回

    FA = "bestvideo+bestaudio/best"
    FB = "bestvideo[height<=?1080]+bestaudio/best"
    FC = "bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best"
    FD = "bestvideo[ext=mp4]+bestaudio[ext=m4a]/best[ext=mp4]/best"

    # 目前的FH的这种LS的condition 可以保证下载的文件不会太大 但是 也可能因为size的限制而下载较低分辨率的文件
    maxsize = ytb_cfg_maxsize_str
    LS = f"[filesize<={maxsize}]"
    FH = f"bestvideo*[ext=mp4]{LS}+bestaudio[ext=m4a]{LS}/best[ext=mp4]{LS}/bestvideo*{LS}+bestaudio{LS}/best{LS}"
    # FI只下载mp4 m4a两种格式的文件 最终会合并到mp4文件
    FI = f"bestvideo*[ext=mp4]{LS}+bestaudio[ext=m4a]{LS}/best[ext=mp4]{LS}"

    req.check()

    AT.assert_(req.data_dir, "err30672 data_dir is bad")

    datadir = DkFile(req.data_dir)
    dir = datadir.pathstr  # '/path/to/mydata'
    fmt = FI
    tot = ytb_cfg_timeout  # seconds
    if True or req.rate_limit > 0:
        ratei = req.rate_limit
        AT.assert_(ratei >= 1000, "err62121 bad rate_limit")
        ratei = int(ratei / 1000)
        rate = f"{ratei}K"

    url = req.url
    proxy = ""  # 如果proxy是空字符串 将直接网络连接 --proxy ""
    proxy = req.proxy  # for example 'socks5://192.168.0.34:7890'

    AT.assert_(req.proxy is not None, "err35925")

    cmd_ytb = r"d:\yt-dlp2023.exe "
    cmd_ytb = r"yt-dlp "
    AT.assert_(
        req.cmd is not None and len(req.cmd) > 0 and req.cmd.endswith(" "), "err86555"
    )
    cmd_ytb = req.cmd

    LN = 150
    LN = 0

    LTitle = 70

    short_pre = ""
    if False and req.ytb_item.has_short():
        short_pre = "[SH]"

    OPT = f"%(title).{LTitle}B{short_pre}[%(id)s].%(ext)s"

    """
    res:720 prefers larger videos, but no larger than 720p and the smallest video if there are no videos less than 720p.
    
    """
    FMS = f"res:{req.ytb_resolution}"

    print_json = "--print-json "
    print_json = " "

    cmd_ytb += f'--ignore-config --socket-timeout {tot} --limit-rate "{rate}" --proxy "{proxy}" '

    RETRIES = 20
    cmd_ytb += f'--retries {RETRIES} '

    if LN:
        """
        似乎 LN与LTitle可能形成冲突？ 比如将 id 或者id的一部分 被trim掉 所以暂时关闭
        
        ct-2024-01-16T01-15-57_r-10001919_ok-no_ytbid-XGzUw+envy0\
            Blue Swimsuit Fashion Show 2⧸4 [AI룩북] AI ART 룩북(Stable Diffusion lookbook) #aiart #ai룩북 #ailookboo[XGzUw_e.mp4
        """
        cmd_ytb += f"--trim-filenames {LN} "

    # 这个 max-filesize 参数为何没有生效？改成M的单位试试看 似乎还是不行
    cmd_ytb += f"--no-progress --part {print_json} --max-filesize {maxsize} "
    cmd_ytb += f"--write-description --write-info-json --write-annotations  "

    # dump-single-json 这个参数什么意义 好像不能在此使用？不会下载文件？
    # cmd += f'--dump-single-json   '

    # 有些视频的comments好像极多 会导致下载出现问题 例如 https://www.youtube.com/watch?v=kYc7D9qGSAU
    # cmd += f'--write-comments  '

    cmd_ytb += f"--no-cache-dir --windows-filenames  "
    cmd_ytb += f"--write-thumbnail  "
    cmd_ytb += f"--write-all-thumbnails  "

    def parse_items_limit() -> int:
        ur = urlparse(req.ytb_item.url_original)
        qr = dict(parse_qsl(ur.query))
        KEY = YtbConf.url_key_maxitems
        r = YtbConf.playlist_maxitems_default
        if KEY in qr and qr[KEY].isnumeric():
            r = int(qr[KEY])
        r = min(max(r, 2), 500)
        return r

    items_limit = parse_items_limit()
    cmd_ytb += f"--write-playlist-metafiles --yes-playlist --playlist-items 1:{items_limit} "
    cmd_ytb += f"--playlist-random "

    """
    -o --output
    -f --format
    -S --format-sort
    
    """
    cmd_ytb += f'--output "{dir}/{OPT}" --format "{fmt}" --format-sort "{FMS}" '

    cmd_ytb += f' "{url}" '

    # if False:
    #     cmd_ytb = f'd:\\yt-dlp2023.exe --proxy "{proxy}" {url} --format bestaudio'

    req.cmd = cmd_ytb
    req.ytb_format = fmt
    start_epoch = AT.fepochSecs()
    req.start_epoch = start_epoch
    iprint_debug(f"cmd: {cmd_ytb}")

    if req.no_download:
        iprint_debug("step65173 no download")
        return None

    if True:
        # 用下划线开头 可以让文件名称顺序在最前面
        Path(join(req.data_dir, f"_r-request_{req.ytb_ytbid.to_str_with_prefix()}.json")).write_text(
            req.to_json(indent=4, ensure_ascii=False), encoding="utf-8"  # type:ignore
        )

    rc, out, err = run_simple_b(cmd_ytb, _debug=True)
    assert isinstance(out, str) and isinstance(err, str)
    end_epoch = AT.fepochSecs()

    files_path = (join(dir, f) for f in os.listdir(dir))
    files = (DkFile(f) for f in files_path)
    exts = YtbConf.media_file_exts_with_dot
    files = [
        f
        for f in files
        if any(f.basename.endswith(e) for e in exts)
        if f.filesize >= YtbConf.media_min_filesize
    ]

    if files:
        return DownResult(
            ok=True,
            rcode=0,
            err=err,
            data_dir=dir,
            url=url,
            req=req,
            stdout=out,
            stderr=err,
            start_epoch=start_epoch,
            end_epoch=end_epoch,
        )

    else:
        return DownResult(
            ok=False,
            rcode=rc,
            err=f"rc={rc} out={out} err={err}",
            data_dir=dir,
            url=url,
            req=req,
            stdout=out,
            stderr=err,
            start_epoch=start_epoch,
            end_epoch=end_epoch,
        )


"""


"""



from __future__ import annotations

import os
from dataclasses import dataclass, field
from os.path import join
from typing import List

from dataclasses_json import dataclass_json
from dknovautils import AT, DkFile

from .common_ctts_ytb import YtbConf
from .down_request import DownRequest, _f_res_version


@dataclass_json
@dataclass
class DownResult:
    version: str = field(default_factory=_f_res_version)
    ok: bool = False
    rcode: int = -1
    data_dir: str = ""
    err: str = ""
    url: str = ""
    res_id: str = ""
    infos: List[str] = field(default_factory=list)
    req: DownRequest = field(default_factory=DownRequest)
    start_epoch: float = 0
    start_epoch_str: str = ""
    end_epoch: float = 0
    end_epoch_str: str = ""
    files_cnt: int = 0
    files_size: int = 0
    files: List[str] = field(default_factory=list)
    rid: int = 0
    stdout: str = ""
    stderr: str = ""
    netv: int = 0  # 下载网速 byte/s
    last_new_urls: int = -1

    @staticmethod
    def from_req(req: DownRequest) -> DownResult:
        r = DownResult(
            data_dir=req.data_dir,
            url=req.url,
            req=req,
        )
        return r

    def check(self) -> bool:
        AT.assert_(
            self.url and self.data_dir and self.req and self.req.check(),
            "err36475 DownResult format error",
        )

        return True

    def after_run(self) -> None:
        """
        包括获取和计算目录内的文件信息
        可以重复调用 不会重复生成数据

        """
        assert self.start_epoch > 0, 'err16785'

        if self.start_epoch_str == "":
            if self.start_epoch > 0:
                self.start_epoch_str = AT.sdf_logger_format_datetime(
                    int(self.start_epoch * 1000)
                )
            if self.end_epoch > 0:
                self.end_epoch_str = AT.sdf_logger_format_datetime(
                    int(self.end_epoch * 1000)
                )

            # fs=[DkFile(f) for f in os.listdir(self.data_dir)]
            fs = DkFile.listdir(self.data_dir)
            self.files_cnt = len(fs)
            self.files_size = sum(f.filesize for f in fs)
            self.files = []
            for f in fs:
                if f.filesize < YtbConf.media_min_filesize:
                    continue
                mtime = os.path.getmtime(f.pathstr)
                dtstr = AT.sdf_logger_format_datetime(int(mtime * 1000))
                ln = [str(f.filesize), dtstr, f.basename]
                ln = ",".join(ln)
                self.files.append(ln)

            if self.ok:
                self.err = ""

            if self.err:
                self.err = self.err[:100]

            if self.stdout:
                self.stdout = self.stdout[:1000]

            if self.stderr:
                self.stderr = self.stderr[:1000]

            # 计算平均网速
            res = self
            span = res.end_epoch - res.start_epoch
            netv = 0 if not span else int(res.files_size / span)
            res.netv = netv

            self.check()

    def write_json_file(self) -> None:
        # 在目录下写入 result.json 文件 文件名称用下划线开头 可以让文件顺序排在前面
        res = self
        DkFile(join(res.data_dir, "_result.json")).path.write_text(
            res.to_json(indent=4, ensure_ascii=False), encoding="utf-8"  # type:ignore
        )

    def cal_title_parts_size_lmt(self) -> int:
        bsize = len(self.cal_sdownload_dir_name(ctitle=''))
        MAXbytes = 240
        return MAXbytes - bsize
        # astr = f'ct-2023-01-01_r-1111111_ok-yes_ytdip-{self.req.ytb_ytbid.id_original}'

    def cal_sdownload_dir_name(self, ctitle: str | None = None) -> str:
        res = self
        ctstr = AT.sdf_logger_format_datetime(dt=int(self.start_epoch * 1000), precise='d', noColon=True)
        ttp = '' if ctitle is None else f'_tt-{ctitle}'
        dirname = (
            f'ct-{ctstr}'
            f'_r-{self.rid}_ok-{"yes" if res.ok else "no"}'
            f'{ttp}'
            f"_{self.req.ytb_ytbid.to_str_with_prefix()}"
        )
        return dirname

    #             bname_7z = f"r-{self.rid}_{req.ytb_ytbid.to_str_with_prefix()}.{self.fkey}.7z"
    def cal_s7z_partial_name(self, p7z_vn: int, secret: str, fkey: str) -> str:
        assert 0 <= p7z_vn < 1000
        ctstr = AT.sdf_logger_format_datetime(dt=int(self.start_epoch * 1000), precise='d', noColon=True)
        phstr = AT.md5_hex(secret.encode())[:4]
        bname_7z = (
            f"ct-{ctstr}"
            f"_r-{self.rid}_{self.req.ytb_ytbid.to_str_with_prefix()}"
            f"_ph-{phstr}_vn-{str(p7z_vn + 1000)[1:]}"  # 为了方便替换 vn必须在最后一部分
            f".{fkey}.7z"
        )
        return bname_7z

    def replace_s7z_vn(self, oldfilename: str, p7z_vn: int) -> str:
        """replace vn to new value """
        assert 0 <= p7z_vn < 1000
        fnpa, *fnps = oldfilename.split('.')
        *bpa, _ = fnpa.split('_')
        fname2 = '.'.join(['_'.join([*bpa, f'vn-{str(p7z_vn + 1000)[1:]}']), *fnps])
        return fname2

    def to_str_simple(self) -> str:
        self.check()
        m = {
            "ok": self.ok,
            "rcode": self.rcode,
            "err": self.err,
            "url": self.url,
            "req.uuid": self.req.uuid,
        }
        return str(m)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Protocol

from dataclasses_json import dataclass_json
from dknovautils import AT

from .entity import YtbId, YtbItem
from .common_etc import AT2

_req_version = "0.2.1"


def _f_req_version() -> str:
    return f'DRQ-{_req_version}'


def _f_res_version() -> str:
    return f'DRS-{_req_version}'


@dataclass_json
@dataclass
class DownRequest:
    version: str = field(default_factory=_f_req_version)
    url: str = ""
    domain: str = ""
    resource_key: str = ""
    data_dir: str = ""
    uuid: str = ""
    start_epoch: float = 0
    rate_limit: int = 1000 * 1000
    ytb_resolution: int = 720
    ytb_format: str = ""
    ytb_ytbid: YtbId = field(default_factory=YtbId)
    cmd: str = ""
    proxy: str = ""
    no_download: bool = False  # 是否dry-run 并不真实的下载数据
    source: str = ""
    ytb_item: YtbItem = field(default_factory=YtbItem)

    def init(self) -> None:
        self.uuid = AT2.gen_uuid_str()

    def check(self) -> bool:
        AT.assert_(
            self.url and self.uuid and self.start_epoch >= 0,
            f"err40638 {self.to_json()}",  # type:ignore
        )
        AT.assert_(self.domain and self.resource_key, "err57667")
        AT.assert_(not self.ytb_item.is_empty(), "err78243")
        return True


class IDownRequest(Protocol):
    pass



from __future__ import annotations

import re
from typing import Dict, Iterable, Iterator, Callable, Set, cast
from urllib.parse import urlparse, parse_qsl, urlencode

from dknovautils import AT, DkFile, iprint_warn, iprint_info, iprint_debug

from .common_util import CommonUtil, ln_filter_nor_comments_nor_empty
from .down_result import DownResult
from .entity import YtbId, IYtbItem, YtbItem
from .common_ctts_ytb import ytb_url_prefix_www_youtube_com, ytb_code_len_a_11, ytb_reg_simple_code11, \
    YtbConf, \
    ytb_reg_simple_shorts, ytb_url_prefix_www_youtube_com_watch, ytb_reg_simple_id_all, ytb_reg_simple_myid_all
from .common_etc import DkFile2, AT2, f_simple_limiter

from pyrate_limiter import Duration, Rate, Limiter
from flask import Flask, request
import flask

from .common_imports import  *

class YtbUtil:






    @staticmethod
    def ytb_check_is_myid(id: str) -> bool:
        assert re.fullmatch(ytb_reg_simple_myid_all, id), 'err95031'
        return "_" not in id

    @staticmethod
    def ytb_check_is_original_id(id: str) -> bool:
        assert re.fullmatch(ytb_reg_simple_id_all, id), 'err95031'
        return "+" not in id

    @staticmethod
    def ytb_original_to_myid(id: str) -> str:
        AT.assert_(YtbUtil.ytb_check_is_original_id(id))
        return id.replace("_", "+")

    @staticmethod
    def ytb_myid_to_original(id: str) -> str:
        AT.assert_(YtbUtil.ytb_check_is_myid(id))
        return id.replace("+", "_")

    @staticmethod
    def ytb_item_check(ytbitem: YtbItem) -> None:
        assert not ytbitem.is_empty(), "err61712"
        url = ytbitem.url
        assert url and len(url) > 10, f"err33528 bad url {url}"
        assert url.startswith(ytb_url_prefix_www_youtube_com), \
            f"err23980 bad url {url}"

        assert 1 == int(ytbitem.has_list()) + int(ytbitem.has_vcode()) + int(ytbitem.has_short()), 'err36996'

        if ytbitem.vcode:
            assert len(ytbitem.vcode) == ytb_code_len_a_11, "err84820 bad v"

        if ytbitem.playlist:
            # list_demo_len = len("PLBSs8pc_eAe5rzljfbZ4qmhQK8eb56pRX")
            _demo_a = 'PLWKjhJtqVAblStefaz+YOVpDWqcRScc2s'
            assert len(_demo_a) == 34
            assert ytb_code_len_a_11 < len(ytbitem.playlist) < len(_demo_a) * 2, \
                f'err78746 length error "{ytbitem.playlist}" '

        if ytbitem.short:
            assert len(ytbitem.short) == ytb_code_len_a_11, "err30341 bad shorts"

    @classmethod
    def parse_ytbid_vcode_from_downloaded_filename(cls, dkf: DkFile | str) -> YtbId | None:
        basename = dkf.basename if isinstance(dkf, DkFile) else dkf
        if not basename.endswith('.mp4'):
            return None
        bname = basename[:-4]  # 去掉 .mp4 字符
        bname = bname[-(ytb_code_len_a_11 + 2):]
        if len(bname) < ytb_code_len_a_11 + 2:
            return None
        # s, *css, e = bname[-(11 + 2):] 不建议对str使用 destructure
        s, css, e = bname[0], bname[1:-1], bname[-1]
        oid = css  # "".join(css)
        ok = (s == '[' and e == ']' and re.fullmatch(ytb_reg_simple_code11, oid))

        if not ok:
            return None
        else:
            r = YtbId(
                prefix=YtbConf.IdPrefix_vcode_ytbid,
                myid=YtbUtil.ytb_original_to_myid(oid),
                id_original=oid)
            return r

    @staticmethod
    def parse_ytbitem_from_url(url: str, silent: bool = False) -> IYtbItem | None:
        """
        支持url中的 v list short 两个参数 其他参数都自动删除

        两个参数不能同时存在 如果同时存在 则作为list处理

        https://www.youtube.com/watch?v=__xx_-xxxxx
        https://www.youtube.com/watch?list=xxxxdfsfsfsdfsfsdfsf
        https://www.youtube.com/watch?v=&list=xxxxdfsfsfsdfsfsdfsf

        https://www.youtube.com/shorts/BrkK9WrcRSQ
        https://www.youtube.com/shorts/BrkK9WrcRSQ?feature=share

        算法

        if url.startswith('xx/shorts'
            short
        elif url.startswith('xx/watch'
            v or list

        """
        try:
            ur = urlparse(url)
            qr = dict(parse_qsl(ur.query))

            def f_url_fullpath(url: str) -> str:
                ur = urlparse(url)
                url2 = ur._replace(query="", fragment="").geturl()
                return url2

            url_full_path = f_url_fullpath(url)
            watch_fullpath = ytb_url_prefix_www_youtube_com_watch == url_full_path

            def f_url2(qr: Dict[str, str]) -> str:
                # [qr.pop(k) for k in list(qr.keys()) if k not in ("v", "list")]
                query2 = urlencode(qr)
                url2 = ur._replace(query=query2, fragment="").geturl()
                return url2

            if re.fullmatch(ytb_reg_simple_shorts, url_full_path):
                scode = ur.path.split('/')[-1]
                assert len(scode) == ytb_code_len_a_11, 'err20039'
                assert YtbConf.short2vcode_将短视频url转换为普通url(), 'err87260'
                ytbitem = YtbItem(
                    url=f"{ytb_url_prefix_www_youtube_com_watch}?v={scode}",
                    vcode=scode,
                    netloc=ur.netloc,
                    url_original=url,
                )
            elif watch_fullpath and qr.get('v', '') and not qr.get('list', ''):
                ytbitem = YtbItem(
                    url=f_url2({'v': qr['v']}),
                    vcode=qr['v'],
                    netloc=ur.netloc,
                    url_original=url,
                )
            elif watch_fullpath and qr.get('v', '') and qr.get('list', ''):
                """
                ?v=tttt&list=xxxx
                """
                url2 = f_url2({'list': qr['list'], 'v': qr['v']})
                ytbitem = YtbItem(
                    url=url2,
                    playlist=qr['list'],
                    netloc=ur.netloc,
                    url_original=url,
                )
            else:
                if not silent:
                    iprint_info(f'err21265 bad url {url} {ur}')
                assert False, 'err57071'

            YtbUtil.ytb_item_check(ytbitem)

        except Exception as e:
            if not silent:
                iprint_info(f"err64022 bad url: {url}")
                iprint_info(e)
            return None

        else:
            return ytbitem

    @staticmethod
    def parse_ytbid_from_url(url: str) -> YtbId | None:
        """
        https://www.youtube.com/watch?v=GVl1E-3FxoQ
        https://www.youtube.com/watch?list=La1CTsdfsdfsdfjpoE_8

        """
        ytbitem = YtbUtil.parse_ytbitem_from_url(url)
        if not ytbitem:
            return None
        assert ytbitem
        ytbid = ytbitem.para_value_to_ytbid()
        return ytbid

    @staticmethod
    def parse_ytbid_from_lines(lns: Iterable[str]) -> Iterator[YtbId]:
        """从多行文本中提取信息

            SA
                去掉注释
                strip not empty

                SB
                    not '.mp4' end
                    start https_ytb_com

                    ytb url
                    get ytbid


                SC
                    not '.mp4' end
                    名称无.符号
                    not basename 可以解析dict并且包含 ct r ok
                    匹配id正则表达式
                    len>=11

                    len=11 vcode_original
                    len>11 list_original


                SD
                    来自于 下载数据生成的 目录名称 和 文件名称

                    SDD
                        not '.mp4' end
                        名称无.符号
                        basename 可以解析dict并且包含 ct r ok

                        get ytbid

                    SDF
                        '.mp4' end


                        get ytbid

        """

        tprint = AT2.tprint_fun(verbose=False, use_print=True)

        def f_has_folder_keys(ln: DkFile) -> bool:
            return {'ct', 'r', 'ok'}.issubset(CommonUtil.parse_filename_dict(ln, silent=True).keys())

        tprint(f'step45971')

        def handle_ln(ln: str) -> YtbId | None:
            tprint(ln)

            try:
                SA_all = True

                if SB_https := (
                        not ln.endswith('.mp4') \
                        and ln.startswith(ytb_url_prefix_www_youtube_com)
                ):
                    return YtbUtil.parse_ytbid_from_url(ln)
                elif SC_codes := (
                        not ln.endswith('.mp4') \
                        and '.' not in ln \
                        and not f_has_folder_keys(DkFile(ln)) \
                        and re.fullmatch(ytb_reg_simple_id_all, ln) \
                        and len(ln) >= ytb_code_len_a_11
                ):
                    tprint('step15426')
                    YtbUtil.ytb_check_is_original_id(ln)
                    return YtbId(
                        prefix=YtbConf.IdPrefix_vcode_ytbid if len(ln) == ytb_code_len_a_11 \
                            else YtbConf.IdPrefix_list_ytbip,
                        myid=YtbUtil.ytb_original_to_myid(ln),
                        id_original=ln)
                elif SDD_foldername := (
                        not ln.endswith('.mp4') \
                        and '.' not in ln \
                        and f_has_folder_keys(DkFile(ln))
                ):
                    fndc = CommonUtil.parse_filename_dict(DkFile(ln))
                    k = [k for k in fndc.keys() if k in YtbConf.IdPrefixList][0]
                    vstr = f'{k}-{fndc[k]}'
                    return YtbId.build_from_vstr(vstr)
                elif SDF_filename := (
                        ln.endswith('.mp4')
                ):
                    return YtbUtil.parse_ytbid_vcode_from_downloaded_filename(DkFile(ln))
                else:
                    return None

            except Exception as e:
                tprint(e)
                return None

        lns = ln_filter_nor_comments_nor_empty(lns)
        rit = (r for ln in lns if (r := handle_ln(ln)))
        return rit

    @staticmethod
    def cal_title_part_from_description_filename(res: DownResult, down_dir: DkFile) -> str | None:
        """
        windows文件名暂时不管
linux文件名最多255bytes
所以 综合按照不超过255字节计算

        字节数量<=240字节计算

        至少存在一个 [xxx].des

102S202 宗教哲學[PLCX-BLZ1hDpCrvRshzjwJllCfqvW9k10n].description

        title长度的确定
            替换._符号
            ct-20230101_r-1111111_ok-yes_ytdip-xxxx

        另外一个可以参考的工具是
        pip install python-slugify

        https://stackoverflow.com/questions/295135/turn-a-string-into-a-valid-filename


        """

        import itertools as ite
        def f_trim_by_accumulate(s: str, mlmt: int, acc_len: Callable[[str], int]) -> str:
            assert mlmt > 0
            sa = (acc_len(c) for c in s)
            b = ite.accumulate(sa, lambda a, b: a + b)
            b = (i for i, ac in enumerate(b) if ac > mlmt)
            b = list(ite.islice(b, 1))
            i = b[0] if b else None
            if i is None:
                return s
            else:
                return s[:i]

        lmt_bytes = res.cal_title_parts_size_lmt()
        assert lmt_bytes > 0

        end_part = f'[{res.req.ytb_ytbid.id_original}].description'
        iprint_debug(f'step96729 {end_part}')
        df = [df for df in DkFile2.oswalk_simple_a(down_dir, only_file=True)
              if df.basename.endswith(end_part)]
        if not df:
            return None
        df = df[0]
        name_part = df.basename[:-len(end_part)]
        name_part = (name_part.replace(".", "-")  # 替换掉点好 跟文件名好区别
                     .replace("_", "-"))  # 目录名中不能出现下划线
        iprint_debug(f'step65880 {name_part}')
        name_part = f_trim_by_accumulate(name_part, lmt_bytes, acc_len=lambda c: len(c.encode('utf-8')))
        max_screen_width = 25 * 2
        name_part = f_trim_by_accumulate(name_part, max_screen_width, acc_len=lambda c: 1 if c.isascii() else 2)
        r = name_part
        assert len(r.encode('utf-8')) <= lmt_bytes, 'err57430'
        return r


    @staticmethod
    def write_http_resp_a(
            data: bytes | DkFile,
            filename: str,
            content_type: str | None = None,
            content_length: bool = False,
            attachment: bool = True,
            mock_limit: int | None = None,  # 模拟流控
            lower_break_limit: int | None = None,  # 速度低于限制的时候主动停止。
    ) -> Any:
        """
        写这么复杂 是支持两个方面

        模拟限流的效果。服务端进行流控，模拟低速网络。
        检测到超低的下载速度的时候，服务端可以主动停止传送。

        """

        tprint = AT2.tprint_fun(verbose=False)

        tprint("step36815")
        is_file = True if isinstance(data, DkFile) else False
        is_bytes = True if isinstance(data, bytes) else False
        AT.assert_(is_file or is_bytes)

        limiter, tksize = f_simple_limiter(mock_limit) if mock_limit else (None, None)
        lndata = len(cast(bytes, data)) if is_bytes else cast(DkFile, data).filesize

        if not content_type:
            content_type = CommonUtil.MIME_Stream

        resp_paras = {
            "Content-Type": content_type,
            "Content-Length": lndata,
            "Content-Disposition": f'attachment; filename="{filename}"',
        }
        # 不提供length header
        if not content_length:
            del resp_paras["Content-Length"]

        if not attachment:
            del resp_paras["Content-Disposition"]

        start_epoch = AT.fepochSecs()
        sumsize = 0

        time.sleep(0.01)

        def f_too_slow() -> bool:
            SLMT = 30
            if not lower_break_limit:
                return False
            span = AT.fepochSecs() - start_epoch
            if span >= SLMT and sumsize / span < lower_break_limit:
                # todo 记录业务日志
                tprint("step72347 too slow")
                return True
            else:
                return False

        BSIZE = 1024 * 4 * 10

        def f_gen_data_bytes(data: bytes) -> Any:
            nonlocal sumsize
            tprint(f"step99056")
            Step = tksize if tksize else BSIZE
            AT.assert_(Step >= 1, "err97182")
            lndata = len(data)

            i = 0
            while True:
                if i >= lndata:
                    break
                n = min(Step, lndata - i)
                tprint("step96681")
                if limiter:
                    limiter.try_acquire("token")
                r: bytes = data[i: (i + n)]
                sumsize += n
                if f_too_slow():
                    return
                yield r
                tprint(f"step11918 send data {n}")
                i += n

        def f_gen_data_dkfile(data: DkFile) -> Any:
            nonlocal sumsize
            tprint("step65639")
            data: DkFile = data
            rsize = tksize if tksize else BSIZE
            with open(data.pathstr, "rb") as file:
                fi = io.FileIO(file.fileno())
                fb = io.BufferedReader(fi)
                while True:
                    if limiter:
                        limiter.try_acquire("token")
                    r: bytes = fb.read(rsize)
                    if not r:
                        break
                    sumsize += len(r)
                    if f_too_slow():
                        return
                    tprint(f"step33520 yield {len(r)}")
                    yield r

        if isinstance(data, bytes):
            return f_gen_data_bytes(data), resp_paras
        else:
            return f_gen_data_dkfile(data), resp_paras



def cal_v_width(s: str) -> int:
    # non ascii =2 width on screen
    r = sum((1 if c.isascii() else 2) for c in s)
    return r



from __future__ import annotations

from dataclasses import dataclass

from .common_etc import AT2


@dataclass
class DownloadParas:
    """
    该类后续会废弃

    """
    epoch: int = 0  # epoch ms
    salt: int = 0
    fid: int = 0
    hash: str = ""

    def join_str(self) -> str:
        sa = f"epoch={self.epoch}&salt={self.salt}&fid={self.fid}"
        if self.hash:
            sa += f"&hash={self.hash}"
        return sa

    def gen_hash(self, secret: str) -> None:
        def _h(s: str) -> str:
            return AT2.md5_hex(bytes(s, "utf-8"))

        self.hash = ""
        self.hash = _h(_h(self.join_str()) + secret)[:6]



from __future__ import annotations

from takahom.common.common_util import *
from takahom.common.down_paras import DownloadParas
from takahom.common.down_result import DownResult
from takahom.common.common_etc import AT2
from takahom.common.ytb_util import YtbUtil


@dataclass_json
@dataclass
class DLClientState:
    results: List[DownResult] = field(default_factory=list)
    fid: int = 0
    rid: int = 0


class DownloadClient(object):
    def __init__(
            self,
            fkey: str = 'TAK',
            secret: str = 'mysecret',
            host: str = "localhost",
            port: int = 8800,
            client_port: int = 8700,
            work_dir: DkFile = DkFile(r"d:\ytb_download_client"),
    ) -> None:
        self.fkey = fkey

        self.secret = secret

        self.TA = 30
        self.TB = 60

        # self.server_port = server_port

        self.server_host = host
        self.server_port = port
        self.server_url = f"http://{host}:{port}"

        self.client_port = client_port

        # self.test_port = 8000

        # headless模式下 firefox运行会出错 可能使用profile会改变 今后再说
        self.browser_headless = False
        # self.browser_headless = True
        self.sleep_a = 300
        self.sleep_b = 30

        self.empty_postfix = ".empty"

        # 平均网速的下限值
        self.lower_lmt_rate = 10 * 1024

        self.work_dir = work_dir

        self.fid = 10 ** 5

        if True:
            # self.fkey = "SEL"
            self.TA = 3
            self.TB = 10

            self.sleep_a = 30
            self.sleep_b = 5
            self.lower_lmt_rate = 3 * 1024

        AT.assert_(
            self.work_dir.exists() and self.work_dir.is_dir(),
            f"err51563 dir not exists {self.work_dir.path}",
        )

        for d in ["cresp", "cdownload", "ctest", "c7z"]:
            os.makedirs(join(self.work_dir.path, d), exist_ok=True)

        # 用目录下的文件 扫描 然后初始化
        fids = [
            int(self.br_parse_filename(DkFile(fname))["fid"])
            for fname in os.listdir(join(self.work_dir.path, "cresp"))
        ]
        if fids:
            self.fid = max(fids) + 1
            AT.assert_(self.fid > 0, "err30212")

        if True:
            self.f_check_hello()

    def f_check_hello(self) -> None:
        start = AT.fepochSecs()

        def fhello() -> None:
            status, data = self.f_http_get(self.server_host, self.server_port, "/hello")
            AT.assert_(status == 200 and "ALLOK" in data)

        ts = []
        for i in range(100):
            thread = threading.Thread(target=fhello, args=())
            thread.start()
            ts.append(thread)

        for t in ts:
            t.join()

        end = AT.fepochSecs()
        iprint_info(f"f_test_hello end {end - start}")

    def f_http_get(self, host: str, port: int, url: str) -> Tuple[int, str]:
        import http.client

        conn = http.client.HTTPConnection(host, port)
        conn.request("GET", url)
        r1 = conn.getresponse()
        # print(r1.status, r1.reason)
        # 200 OK
        data1 = r1.read()  # This will return entire content.
        # print(data1)
        # The following example demonstrates reading data in chunks.
        conn.close()
        return r1.status, str(data1, "utf-8")

    def start(self) -> None:
        thread = threading.Thread(target=self.log_job, args=(), daemon=True)
        thread.start()

        if False:
            thread = threading.Thread(target=self.main_job, args=(), daemon=False)
            thread.start()

        self.main_job()

        if False:
            self._start_test_server()

        pass

    def br_run(
            self, *, paras: DownloadParas, url: str, headless: bool = False
    ) -> Tuple[DkFile | None, str]:
        path_firefox = r"C:\Program Files\Mozilla Firefox\firefox.exe"
        cmd = f'"{path_firefox}" '
        cmd = f'"{path_firefox}" -p downloader --no-remote '
        cmd = f'"{path_firefox}" -no-remote -P downloader '
        cmd = f'"{path_firefox}" -no-remote -P downloader -new-instance '
        cmd = f'"{path_firefox}" -P downloader '

        #  -no-remote -P downloader -new-instance
        if headless:
            cmd += "-headless "
        cmd += f'-url "{url}" '
        DKProcessUtil.run_simple_a(cmd)

        def fcheck() -> Tuple[DkFile | None, str]:
            TA = self.TA
            TB = self.TB

            start_download = AT.fepochSecs()

            time.sleep(TA)

            while True:
                try:
                    now = AT.fepochSecs()
                    files = self.br_list_all_files()

                    # if files ==[] or not self.br_is_downloading():
                    if files == []:
                        # 下载出错了
                        return None, "err34683 no files"

                    file = self.br_file_downloading(paras)

                    low_rate = 1024 * 5
                    low_rate = self.lower_lmt_rate
                    span_lmt = 60
                    rate速度太低 = (
                        None
                        if file is None
                        else file.filesize / (now - start_download) < low_rate
                    )
                    utime_更新超时 = (
                        None
                        if file is None
                        else now - os.path.getmtime(file.path) > 60 * 3
                    )
                    if (
                            file is not None
                            and now - start_download >= span_lmt
                            and (rate速度太低 or utime_更新超时)
                    ):
                        # 显然firefox已经超时
                        iprint_info(
                            f"err36331 utimebad {utime_更新超时} ratebad {rate速度太低}"
                        )
                        return None, CommonUtil.Err_Timeout

                    if file is not None:
                        continue

                    def f_ok(file: DkFile) -> bool:
                        r = str(paras.epoch) in file.basename
                        r2 = not self.br_is_partial_file(file)
                        return r and r2

                    files = [file for file in files if f_ok(file)]

                    AT.assert_(len(files) == 1, "err54281")
                    file = files[0]
                    dc = self.br_parse_filename(file)

                    dcok = dc["ln"] and dc["md5"] and dc["epoch"]
                    AT.assert_(dcok, "err51148")

                    def md5ok() -> bool:
                        m = dc["md5"]
                        AT.assert_(len(dc["md5"]) == 8, "err12586")
                        r = m == (DkFile.file_md5(file.pathstr)[:8])
                        return r

                    mtime = os.path.getmtime(file.path)
                    ln = int(dc["ln"])
                    # 这个逻辑 取决于mtime是否被firefox正确的设置了
                    if mtime < now - TA and file.filesize == ln and md5ok():
                        # file ok ,return
                        return file, ""
                    elif mtime < now - TB and (file.filesize != ln or not md5ok()):
                        # bad, return 可能是网络中断 或者服务器提前终止传送
                        return None, "err82641 err data"
                    else:
                        pass
                finally:
                    time.sleep(TA)
                    pass

        r = fcheck()

        return r

    def br_list_all_files(self) -> List[DkFile]:
        # 用 fkey 过滤 按照ctime顺序排列 todo
        # 包括 partial文件
        dir = "path/to/dir"
        dir = r"C:\Users\dknova\Downloads"
        fkey = "." + self.fkey + "."
        fs = [DkFile(join(dir, fname)) for fname in os.listdir(dir) if fkey in fname]
        return fs

    def br_is_partial_file(self, df: DkFile) -> bool:
        return df.basename.endswith(".part")

    def br_file_downloading(self, paras: DownloadParas) -> DkFile | None:
        fs = self.br_list_all_files()
        fs = [
            file
            for file in fs
            if str(paras.epoch) in file.basename and self.br_is_partial_file(file)
        ]
        r = len(fs) > 0
        if r:
            return fs[0]
        else:
            return None

    def br_kill_process(self) -> None:
        os.system('wmic process where name="firefox.exe" delete')
        # 查看进程列表  wmic process where name="firefox.exe" list

    def br_parse_filename(self, df: DkFile) -> Dict[str, str]:
        # 解析正常文件名称 以及 partial文件名
        return CommonUtil.parse_filename_dict(df, check=True)

    def log_job(self) -> None:
        pass

    def _subdir(self, d: str) -> DkFile:
        r = DkFile(join(self.work_dir.path, d))
        AT.assert_(r.is_dir() and r.exists(), f"err57249 {r.path}")
        return r
        pass

    def main_job(self) -> None:
        def f_is_test_url(url: str) -> bool:
            return url.startswith(f"{self.server_url}/test")

        def f_is_data_url(url: str) -> bool:
            return url.startswith(f"{self.server_url}/data")

        time.sleep(3)
        cnt_a = 0
        while True:
            if False:
                self.br_kill_process()
                ta = 1.0
                ta = 5.0
                ta = 10.0
                time.sleep(ta)

            def fa() -> None:
                for file in self.br_list_all_files():
                    if not self.br_is_partial_file(file):
                        os.remove(file.pathstr)
                        iprint_debug(f"remove file:{file.pathstr}")

            fa()

            AT.assert_(
                [
                    file
                    for file in self.br_list_all_files()
                    if not self.br_is_partial_file(file)
                ]
                == [],
                "err84362",
            )

            p_mock_limit = 0.2
            vlmt = random.randint(1024, 1024 * 100)
            salt = (
                vlmt if random.uniform(0, 1) < p_mock_limit else random.randint(1, 100)
            )

            paras = DownloadParas(epoch=AT.fepochMillis(), salt=salt, fid=self.fid)

            paras.gen_hash(self.secret)

            # url = f'http://localhost:{self.test_port}/test_100'+'?'+ paras.join_str()

            # url = f'{self.server_url}/test'+'?'+ paras.join_str()
            url = f"{self.server_url}/data" + "?" + paras.join_str()

            rrun = self.br_run(
                paras=paras, url=url, headless=self.browser_headless
            )

            file: DkFile | None = rrun[0]
            err = rrun[1]

            iprint_debug(
                f'get file {cnt_a} {"ok" if file is not None else "err"}. browser file:{file.path if file is not None else None} err:{err}'
            )

            v下载正常 = file is not None
            v数据empty = v下载正常 and file and file.basename.endswith(self.empty_postfix)
            v超时 = file is None and err == CommonUtil.Err_Timeout
            v下载不正常 = file is None

            if v下载正常 and v数据empty:
                # 将empty文件删除 记录日志
                assert file
                os.remove(file.path)
                iprint_debug("empty file removed")

            elif v下载正常 and not v数据empty:
                assert file
                dc = self.br_parse_filename(file)
                dir_ctest = self._subdir("ctest")
                dir_cresp = self._subdir("cresp")
                dir_c7z = self._subdir("c7z")
                dir_cdownload = self._subdir("cdownload")

                if False:
                    AT.never()
                elif f_is_test_url(url):
                    AT.assert_(dc["code"] == "110")
                    # move data to ctest 目录
                    shutil.move(file.pathstr, join(dir_ctest.pathstr, file.basename))

                elif f_is_data_url(url):
                    # 将下载的文件移动到 cresp
                    def fcheck(file: DkFile) -> None:
                        """
                                           v是7z分片压缩的格式 =
                        不是 .empty 扩展名
                        t-7z r-123 n-3 i-1
                        """
                        AT.assert_(
                            dc["code"] == "100"
                            and not file.basename.endswith(".empty")
                            and dc["t"] == "7z"
                            and dc["r"]
                            and dc["n"]
                            and dc["i"]
                            and True,
                            "err26942 bad filename format",
                        )

                    fcheck(file)

                    shutil.move(file.pathstr, join(dir_cresp.pathstr, file.basename))

                    i = int(dc["i"])
                    n = int(dc["n"])
                    rid = int(dc["r"])

                    dest7zname = f"r-{rid}_n-{n}.7z.{str(1000 + i)[1:]}"
                    dest7zpath = join(dir_c7z.pathstr, dest7zname)
                    shutil.copy(join(dir_cresp.pathstr, file.basename), dest7zpath)

                    v获取R中的最后一个7z文件 = n == i
                    if v获取R中的最后一个7z文件:
                        path_7z_exe = r"C:\Program Files\7-Zip\7z.exe"
                        destdir = join(dir_cdownload.pathstr, f"r-{rid}")
                        os.makedirs(destdir, exist_ok=False)
                        S = os.sep
                        pwd = self.secret
                        fname7z = f"r-{rid}_n-{n}.7z.001"
                        cmd = f'"{path_7z_exe}" x '
                        # cmd += f'-p{pwd} -o"{destdir}" "{fname7z}" '
                        cmd += f'-p{pwd} -o"{dir_cdownload.pathstr}" "{fname7z}" '
                        run_simple_a(cmd)

                        fs = DkFile.listdir(destdir)
                        AT.assert_(len(fs) > 0, "err91044")

                    AT.assert_(dc["fid"], "err53720")

                    self.fid = int(dc["fid"]) + 1

                    # 客户端可以根据下载的数据，进行大致的流控 todo

                    pass
                else:
                    AT.never()

                if False:
                    pass

            elif v下载不正常:
                # 记录日志
                iprint_warn(f"err15555 {err}")
                pass

            else:
                AT.never()

            if (v下载正常 and v数据empty) or v下载不正常:
                time.sleep(self.sleep_a)
            else:
                time.sleep(self.sleep_b)

            cnt_a += 1

    def _start_test_server(self) -> None:
        from flask import Flask, request

        app = Flask(__name__)

        @app.route("/hello")
        def f_hello() -> Any:
            return "Hello, World! allok "

        @app.route("/test")
        def f_test() -> Any:
            """
                    ?epoch=123231&salt=23123123&fid=13213123&hash=Vhash
            Vhash = hash(hash('epoch=123231&salt=23123123&id=13213123')+ SECRET)

            """

            demostr = "".join(str(random.randint(1, 9)) for x in range(0, 10 ** 3))
            demostr = demostr * (10 ** 3)
            demobytes = bytes(demostr, "utf-8")
            demomd5 = AT2.md5_hex(demobytes)[:8]

            fidstr = request.args.get("fid", "")
            fid = int(fidstr)

            epoch = int(request.args.get("epoch", ""))

            AT.assert_(fid > 0 and epoch > 0, "err66364")

            filename = f"fid-{fid}_ln-{len(demobytes)}_md5-{demomd5}_epoch-{epoch}.{self.fkey}.data"

            return (YtbUtil.write_http_resp_a(data=demobytes, filename=filename))

        if True:
            _debug = False
            app.run(host="0.0.0.0", port=self.client_port, debug=_debug)

    pass


def main(args: Any) -> None:
    client = DownloadClient(
        host="192.168.0.35", port=8800, work_dir=DkFile(r"d:\ytb_download_client")
    )

    client.start()


if __name__ == "__main__":
    import argparse  # 1

    parser = argparse.ArgumentParser()  # 2

    parser.add_argument("-p", "--profile", type=str, default="dev", help="profile")

    parser.add_argument("--workdir", type=str, help="工作目录")

    parser.add_argument("--port", type=int, default=8800, help="server web port")

    parser.add_argument("--secret", type=str, default="hellobaby", help="7z password")

    parser.add_argument("--location", type=str, default="test", help="location")

    parser.add_argument("--maxcnt", type=int, default=50, help="dircnt")

    parser.add_argument("--reportverbose", type=int, default=30, help="报告的详细程度")

    args = parser.parse_args()  # 4

    main(args)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List

from dataclasses_json import dataclass_json

from takahom.common.down_result import DownResult


@dataclass_json
@dataclass
class DLServerIndex:
    fid: int = 0
    rid: int = 0
    results: List[DownResult] = field(default_factory=list)



try:

    from takahom.common.common_imports import *
    from takahom.common import *

    from pyrate_limiter import Duration, Rate, Limiter

    from flask import Flask, request

    import schedule

    import flask


    from logdecorator import (  # type:ignore
        log_on_start,
        log_on_end,
        log_on_error,
        log_exception,
    )

finally:
    pass



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Protocol

from dataclasses_json import dataclass_json
from dknovautils import AT

from .common_util import CommonUtil
from .common_ctts_ytb import ytb_reg_simple_code11, ytb_reg_simple_id_all, YtbConf


@dataclass_json
@dataclass(eq=True, frozen=True, slots=True)
class YtbId:
    prefix: str = ''  # ytbid
    id_original: str = ''
    myid: str = ''

    def to_str_with_prefix(self) -> str:
        assert self.prefix in YtbConf.IdPrefixList
        assert re.fullmatch(ytb_reg_simple_id_all, self.id_original), 'err44685'
        assert self.prefix and self.id_original and len(self.myid) >= 11, "err92422"
        return f"{self.prefix}-{self.myid}"

    @staticmethod
    def build_from_vstr(vs: str) -> YtbId | None:
        from takahom.common.ytb_util import YtbUtil

        try:
            # 注意 myid中可能包含-符号
            dc = CommonUtil.parse_filename_dict(vs)
            assert len(dc) == 1
            k = list(dc.keys())[0]
            myid = vs[len(k) + 1:]
            assert k in YtbConf.IdPrefixList
            r = YtbId(prefix=k, myid=myid, id_original=YtbUtil.ytb_myid_to_original(myid))
            r.to_str_with_prefix()
            return r
        except Exception as e:
            return None


class IYtbItem(Protocol):
    """
    v list两者只有一个非空

    """

    @property
    def para_vcode(self) -> str | None:
        pass

    @property
    def para_list(self) -> str | None:
        pass

    @property
    def para_short(self) -> str | None:
        pass

    def get_url(self) -> str:
        pass

    def has_vcode(self) -> bool:
        return bool(self.para_vcode)

    def has_list(self) -> bool:
        return bool(self.para_list)

    def has_short(self) -> bool:
        return bool(self.para_short)

    @property
    def para_value(self) -> str:
        """哪个有值就返回哪个"""
        if v := self.para_vcode:
            r = v
        elif v := self.para_list:
            r = v
        elif v := self.para_short:
            r = v
        assert r
        return r

    def para_value_to_ytbid(self) -> YtbId:
        from takahom.common import ytb_util
        id: str = self.para_value
        myid = ytb_util.YtbUtil.ytb_original_to_myid(id)
        prefix = 'ytbid' if self.has_vcode() else 'ytbih' if self.has_short() else 'ytbip'
        return YtbId(myid=myid, prefix=prefix, id_original=id)


@dataclass_json
@dataclass(eq=True, frozen=True, slots=True)
class YtbItem(IYtbItem):
    url: str = ""  # 两个url url_originl可能是不同的。 url可能重组过后的 两者很可能存在差异
    vcode: str | None = None
    playlist: str | None = None
    short: str | None = None
    netloc: str = ""
    url_original: str = ""

    @property
    def para_vcode(self) -> str | None:
        return self.vcode

    @property
    def para_list(self) -> str | None:
        return self.playlist

    @property
    def para_short(self) -> str | None:
        return self.short

    def get_url(self) -> str:
        return self.url

    def is_empty(self) -> bool:
        return self.url == ""



from __future__ import annotations

from dknovautils import *

ytb_url_prefix_www_youtube_com = "https://www.youtube.com/"
ytb_url_prefix_www_youtube_com_watch = "https://www.youtube.com/watch"

ytb_url_demo_moyun = "https://www.youtube.com/watch?v=gf6v59c5yuY"
ytb_url_demo_ydltest = "https://www.youtube.com/watch?v=BaW_jenozKc"

ytb_url_demo100 = ytb_url_demo_moyun
ytb_url_demo100_len = len(ytb_url_demo100)

ytb_code_len_a_11 = 11
assert ytb_code_len_a_11 == len("La1CTjpoE_8")
assert len(ytb_url_prefix_www_youtube_com) > ytb_code_len_a_11

ytb_reg_simple_code11 = re.compile(r"[0-9a-zA-Z-_]{11,11}")
ytb_reg_simple_id_all = re.compile(r"[0-9a-zA-Z-_]{11,90}")
ytb_reg_simple_myid_all = re.compile(r"[0-9a-zA-Z-+]{11,90}")
reg_ytb_url_txt = re.compile(r'ytb_.*_\d{3,4}\.url\.txt')
assert re.fullmatch(reg_ytb_url_txt, 'ytb_erwere_0987.url.txt')
# ytb_xxxx.exd.txt
ytb_reg_exd_files = re.compile(r'ytb_.*\.exd\.txt')
assert re.fullmatch(ytb_reg_exd_files, 'ytb_xxxysfdsf.exd.txt'), 'err342342'

ytb_reg_ytb_txt_files = re.compile(r'ytb_.*\.txt')

assert re.fullmatch(ytb_reg_simple_code11, "__xx_-xxxxx")

ytb_reg_simple_a = re.compile(
    r"https://www\.youtube\.com/watch\?v=[0-9a-zA-Z-_]{11,11}"
)

assert re.fullmatch(
    ytb_reg_simple_a, "https://www.youtube.com/watch?v=__xx_-xxxxx"
)

ytb_reg_simple_shorts = re.compile(
    r"https://www\.youtube\.com/shorts/[0-9a-zA-Z-_]{11,11}"
)

ytb_cfg_maxsize_str = "10.0G"
ytb_cfg_maxsize_str = "5.0G"
ytb_cfg_maxsize_str = "5000M"  # 不确定是否支持G后缀

ytb_domain_name = "youtube"

ytb_cfg_timeout = 120  # secs


class YtbConf:
    IdPrefix_vcode_ytbid = 'ytbid'
    IdPrefix_short_ytbih = 'ytbih'
    IdPrefix_list_ytbip = 'ytbip'

    IdPrefixList = [IdPrefix_vcode_ytbid,
                    IdPrefix_short_ytbih,
                    IdPrefix_list_ytbip]

    playlist_maxitems_default = 100

    url_key_maxitems = '_maxitems_'

    @classproperty
    def media_file_exts(cls) -> Tuple[str, ...]:
        r = ('mp4', 'm4a')
        return r

    @classproperty
    def media_file_exts_with_dot(cls) -> Tuple[str, ...]:
        r = tuple('.' + e for e in cls.media_file_exts)
        return r

    @classproperty
    def media_min_filesize(cls) -> int:
        Min_Filesize = 1024 * 200
        return Min_Filesize

    @classmethod
    def short2vcode_将短视频url转换为普通url(cls) -> bool:
        return True



from __future__ import annotations

from takahom.common.down_request import DownRequest
from takahom.common.common_etc import limit_memory, AT2
from takahom.server.server_base_a import IServerBaseA
# from takahom.server.server_base_a import tgt_email_log
from takahom.server.server_config import ServerConfig
from takahom.server.server_index import DLServerIndex

from takahom.server.server_ytb_util import *


class IServerBaseWeb(IServerBaseA, Protocol):

    def _f_write_file_simple(self, dkfile: DkFile,
                             file_name: str | None = None,
                             attachment: bool = True,
                             content_type: str | None = None,
                             content_length: bool = True) -> Any:

        v130_limit_break = self.sconf.v130_limit_break

        if file_name:
            filename = file_name
        else:
            req_epoch: int = int(request.args.get("epoch", "0"))
            if not req_epoch:
                AT.logger.warn("err61277 no epoch")

            fid = 0
            code = 120
            ln = dkfile.filesize
            datamd5 = AT2.file_md5_hex(dkfile.pathstr)[:8]
            filename = (f"fid-{fid}_code-{code}_epoch-{req_epoch}"
                        f"_ln-{ln}_md5-{datamd5}.{self.fkey}.data")

        vlmt = None

        time.sleep(0.5)
        return YtbUtil.write_http_resp_a(
            data=dkfile,
            filename=filename,
            content_length=content_length,
            content_type=content_type,
            attachment=attachment,
            mock_limit=vlmt,
            lower_break_limit=v130_limit_break,
        )

    def start_web_server(self) -> None:
        from takahom.server import server_base_d

        dir_sresp = self.subdir("sresp")
        dir_s7z = self.subdir("s7z")
        dir_stmp = self.subdir("stmp")

        file_key = self.sconf.file_key
        v130_limit_break = self.sconf.v130_limit_break

        app = Flask(__name__)

        _hello_cnt = 10 ** 5

        @app.route("/hello")
        def f_hello() -> Any:
            nonlocal _hello_cnt
            _hello_cnt += 1
            cnt = _hello_cnt

            cmd = request.args.get("cmd", "")
            if not cmd:
                r = f"Hello, World! ALLOK start:{self.start_app_epoch_str} cnt:{cnt}"
                time.sleep(0.5)
                return r
            elif cmd == "more":
                jstr = server_base_d.tgt_email_log(self, write_email=False)
                return jstr

            elif cmd == "demo_file":
                dkf = DkFile(join(dir_stmp.pathstr, f"hello-{AT2.gen_uuid_str()}"))
                dkf.path.write_bytes(b"hello")
                return self._f_write_file_simple(dkf)

            elif cmd == "index_csv":
                tmp_file = DkFile(join(dir_stmp.path, f"index_{AT.fepochSecs()}.tmp"))
                with open(tmp_file.path, 'w', encoding='utf-8') as file:
                    file.write('#len,md5,filename\n')
                    for dkf in DkFile2.oswalk_simple_a(self.subdir('s7z')):
                        ln = dkf.filesize
                        datamd5 = '0'
                        # datamd5 = AT2.file_md5_hex(dkfile.pathstr)[:8]
                        s = f"{ln},{datamd5},{dkf.basename}\n"
                        file.write(s)
                return self._f_write_file_simple(tmp_file)
            else:
                AT.never()

        @app.route("/file/<string:fname>")
        def f_file(fname: str) -> Any:
            cmd = ''
            if not fname or fname == 'index.html':
                cmd = 'index_file'
            else:
                cmd = 'single_file'

            if cmd == 'index_file':
                # ?att=True 将以附件的形式发送http响应
                attachment = request.args.get("att", False, type=bool)
                tmpfile = DkFile(join(dir_stmp.path, f"index_{AT.fepochSecs()}.html"))

                with open(tmpfile.path, 'w', encoding='utf-8') as file:
                    title = 'index'
                    pre = f'''
                <!DOCTYPE html>
                <html>
                <head>
                <title>{title}</title>
                </head>
                <body>                 
                <pre>   

                                    '''
                    post = '''

                </pre>                    
                </body>
                </html>                       

                                    '''
                    file.write(pre.strip() + '\n')
                    for dkfile in DkFile2.oswalk_simple_a(self.subdir('s7z')):
                        if dkfile.is_dir():
                            continue
                        fname = html.escape(dkfile.basename)
                        fname = dkfile.basename
                        ln = dkfile.filesize
                        datamd5 = '0'
                        p = f'<a href="{fname}">ln={ln},md5={datamd5},file={fname}</a>\n'
                        file.write(p)
                    file.write(post.strip() + '\n')

                return self._f_write_file_simple(
                    tmpfile,
                    file_name='index.html',
                    content_type=CommonUtil.MIME_text_html,
                    attachment=attachment,
                )

            elif cmd == 'single_file':
                AT.assert_(len(fname) > 10, 'err22688')
                file_name = fname
                iprint_debug(f'file_name:{file_name}')
                assert file_name
                dkfile = DkFile2.join_path(dir_s7z, file_name)
                AT.assert_(dkfile.exists(), f'err99421 file not exists {dkfile}')
                return self._f_write_file_simple(dkfile, file_name=file_name)

            else:
                AT.never(cmd)

        @app.route("/data")
        def f_data() -> Any:
            AT.unsupported()

        @app.route("/test")
        def f_test() -> Any:
            """
                    ?epoch=123231&salt=23123123&fid=13213123&hash=Vhash
            Vhash = hash(hash('epoch=123231&salt=23123123&id=13213123')+ SECRET)

            if salt>=1000的时候 用 salt参数作为限制速度 byte/s


            """

            demostr = "".join(str(random.randint(1, 9)) for x in range(0, 10 ** 3))
            demostr = demostr * (10 ** 3)
            demostr = demostr * 10
            demobytes = bytes(demostr, "utf-8")
            demomd5 = AT2.md5_hex(demobytes)[:8]

            fidstr = request.args.get("fid", "")
            fid = int(fidstr)

            epoch = int(request.args.get("epoch", 0))
            AT.assert_(fid > 0 and epoch > 0, "err66364")

            salt = int(request.args.get("salt", 0))
            vlmt = salt if salt >= 1000 else None

            _fid_dummy = 0
            filename = f"fid-{_fid_dummy}_code-110_epoch-{epoch}_ln-{len(demobytes)}_md5-{demomd5}.{file_key}.data"

            iprint_debug(f"step10372")

            return YtbUtil.write_http_resp_a(
                data=demobytes,
                filename=filename,
                mock_limit=vlmt,
                lower_break_limit=v130_limit_break,
            )

        m = {"port": self.port}
        if self.port > 0:
            _debug = False
            iprint_info(f"start app step5 ok. web server start ok. {m} ")
            iprint_info(f"visit web server url: http://localhost:{self.port}/file/index.html ")
            app.run(host="0.0.0.0", port=self.port, debug=_debug)
        else:
            iprint_info(f"start app step5 ok. web server isn't started. {m}")
